"""Test cases for ``Integration.info`` function."""

import typing as t

import pytest
from connector.oai.capability import CapabilityCallableProto, CustomRequest, CustomResponse
from connector.oai.integration import (
    DescriptionData,
    Integration,
    InvalidCapabilityNameError,
    ReservedCapabilityNameError,
)
from connector_sdk_types.generated import (
    AccountStatus,
    Amount,
    BasicCredential,
    CreateAccountRequest,
    CreateAccountResponse,
    CreatedAccount,
    Expense,
    ExpenseApprovalStatus,
    ExpensePaymentStatus,
    ExpenseType,
    ListAccountsRequest,
    ListAccountsResponse,
    ListExpensesRequest,
    ListExpensesResponse,
    NormalizedExpenseApprovalStatus,
    NormalizedExpensePaymentStatus,
    SpendUser,
    StandardCapabilityName,
    Vendor,
)

from .shared_types import (
    AccioRequest,
    AccioResponse,
)

Case: t.TypeAlias = tuple[
    str,
    dict[str, CapabilityCallableProto[t.Any]],
]


def new_integration() -> Integration:
    return Integration(
        app_id="test",
        version="0.1.0",
        auth=BasicCredential,
        exception_handlers=[],
        handle_errors=True,
        description_data=DescriptionData(
            user_friendly_name="testing thing",
            categories=[],
        ),
    )


def case_register_capability_success() -> Case:
    integration = new_integration()
    capability_name = StandardCapabilityName.LIST_ACCOUNTS

    @integration.register_capability(capability_name)
    async def capability(
        args: ListAccountsRequest,
    ) -> ListAccountsResponse:
        return ListAccountsResponse(
            response=[],
            raw_data=None,
        )

    return capability_name.value, integration.capabilities


def case_register_list_expenses_success() -> Case:
    integration = new_integration()
    capability_name = StandardCapabilityName.LIST_EXPENSES

    @integration.register_capability(capability_name)
    async def capability(
        args: ListExpensesRequest,
    ) -> ListExpensesResponse:
        return ListExpensesResponse(
            response=[
                Expense(
                    id="2652b217-d686-4277-aad5-37320e9d9912",
                    report_id="bfc896e8-4bca-4aa4-87df-6341b22ed44d",
                    transaction_date="2024-01-01T:00:00+00:00",
                    payment_date="2024-01-03T:00:00+00:00",
                    total_amount=Amount(
                        amount="100.00",
                        currency="USD",
                    ),
                    paid_amount=Amount(
                        amount="100.00",
                        currency="USD",
                    ),
                    seller=Vendor(
                        id="f2d8d6a7-1a1a-4b3d-a092-936a7b32a0b3",
                        name="Adobe",
                        description="A leading software company known for its creativity products",
                    ),
                    description="Creative Cloud for the Marketing team",
                    type=ExpenseType.REIMBURSEMENT,
                    approval_status=ExpenseApprovalStatus(
                        display_value="Approved",
                        normalized_value=NormalizedExpenseApprovalStatus.APPROVED,
                    ),
                    payment_status=ExpensePaymentStatus(
                        display_value="Paid",
                        normalized_value=NormalizedExpensePaymentStatus.PAID,
                    ),
                    submitter=SpendUser(
                        id="6da0e939-9ba9-497c-bf05-a86e95cbbb49",
                        email="harry@hogwarts.edu",
                    ),
                )
            ],
            raw_data=None,
        )

    return capability_name.value, integration.capabilities


def case_register_capability_with_metadata_success() -> Case:
    integration = new_integration()
    capability_name = StandardCapabilityName.CREATE_ACCOUNT

    @integration.register_capability(
        capability_name,
        display_name="Invite User via Email",
        description="Send an email to invite a user to your organization",
    )
    async def capability(
        args: CreateAccountRequest,
    ) -> CreateAccountResponse:
        return CreateAccountResponse(
            response=CreatedAccount(
                status=AccountStatus.ACTIVE,
                created=True,
            ),
            raw_data=None,
        )

    return capability_name.value, integration.capabilities


def case_register_custom_capability_success() -> Case:
    integration = new_integration()
    capability_name = "accio"

    @integration.register_custom_capability(
        capability_name,
        description="A summoning charm.",
    )
    async def custom_capability(
        args: CustomRequest[AccioRequest],
    ) -> CustomResponse[AccioResponse]:
        return CustomResponse[AccioResponse](
            response=AccioResponse(success=True),
        )

    return capability_name, integration.capabilities


def case_register_custom_capability_reserved_name() -> Case:
    """Don't name your custom capability that! It's ours!"""
    integration = new_integration()
    capability_name = "list_accounts"

    with pytest.raises(ReservedCapabilityNameError):

        @integration.register_custom_capability(
            capability_name,
            description="A summoning charm.",
        )
        async def custom_capability(
            args: CustomRequest[AccioRequest],
        ) -> CustomResponse[AccioResponse]:
            return CustomResponse[AccioResponse](
                response=AccioResponse(success=True),
            )

    return capability_name, integration.capabilities


def case_register_custom_capability_numeric_name() -> Case:
    """We should raise an exception if you give us a non-alpha capability name"""
    integration = new_integration()
    capability_name = "name_with_numb3rs"

    with pytest.raises(InvalidCapabilityNameError) as e:

        @integration.register_custom_capability(
            capability_name,
            description="A summoning charm.",
        )
        async def custom_capability(
            args: CustomRequest[AccioRequest],
        ) -> CustomResponse[AccioResponse]:
            return CustomResponse[AccioResponse](
                response=AccioResponse(success=True),
            )

    assert "Capability names must only contain alphabetic characters and underscores" in str(e)

    return capability_name, integration.capabilities


def case_register_custom_capability_camel_case() -> Case:
    """We should raise an error for a custom capability name that isn't snake_cased"""
    integration = new_integration()
    capability_name = "CamelCaseName"

    with pytest.raises(InvalidCapabilityNameError) as e:

        @integration.register_custom_capability(
            capability_name,
            description="A summoning charm.",
        )
        async def custom_capability(
            args: CustomRequest[AccioRequest],
        ) -> CustomResponse[AccioResponse]:
            return CustomResponse[AccioResponse](
                response=AccioResponse(success=True),
            )

    assert "Capability names must use snake casing" in str(e)

    return capability_name, integration.capabilities



from typing import Any


class CaseInsensitiveDict(dict[str, Any]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.update(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key.lower(), value)

    def __getitem__(self, key):
        return super().__getitem__(key.lower())

    def __delitem__(self, key):
        return super().__delitem__(key.lower())

    def __contains__(self, key):
        return super().__contains__(key.lower())

    def get(self, key, default=None):
        return super().get(key.lower(), default)

    def update(self, other=None, **kwargs):
        if other:
            if hasattr(other, "keys"):
                for k in other:
                    self[k.lower()] = other[k]
            else:
                for k, v in other:
                    self[k.lower()] = v
        for k in kwargs:
            self[k.lower()] = kwargs[k]



import json
import os
import platform
import tarfile
import tempfile
import zipfile
from pathlib import Path
from subprocess import run

import pytest


def test_compile():
    # First, compile from the SDK and check that behavior
    tmpdirname = tempfile.mkdtemp()
    sdk_cli = "connector"
    compile_command = [
        sdk_cli,
        "compile-on-prem",
        "--app-id",
        "mock_connector",
        "--connector-root-module-dir",
        "projects/connectors/python/mock-connector/mock_connector",
        "--output-directory",
        str(tmpdirname),
    ]
    result = run(
        " ".join(compile_command),
        shell=True,
        capture_output=True,
        cwd=Path(os.path.dirname(__file__)).parent.parent.parent.parent.parent,
    )
    stdout = str(result.stdout, "utf-8")
    stderr = str(result.stderr, "utf-8")
    if result.returncode != 0:
        print("stdout:\n", stdout)
        print("stderr:\n", stderr)
        pytest.fail(f"Exited {result.returncode}: {' '.join(compile_command)}")
    archive = stdout.strip(" \n\t")
    assert (
        len(archive) > 0 and len(archive.split(" ")) == 1
    ), f"The output doesn't look like a single file: '{archive}'"
    assert os.path.exists(archive), f"The output isn't a file: '{archive}'"
    with tempfile.TemporaryDirectory() as tempdir:
        if platform.system() == "Windows":
            assert archive.endswith(".zip"), "Archive isn't a zip file"
            with zipfile.ZipFile(archive) as zip:
                zip.extractall(tempdir)
            executable = str(Path(tempdir) / "main.exe")
        else:
            assert archive.endswith(".tar.gz"), "Archive isn't a tar file"
            with tarfile.open(archive) as tar:
                tar.extractall(tempdir)

            executable = str(Path(tempdir) / "main")
            assert os.access(executable, os.X_OK), f"The file isn't executable: '{archive}'"

        # Now, check the compiled connector archive
        info_command = [executable, "info"]
        result = run(
            " ".join(info_command),
            shell=True,
            capture_output=True,
        )
        assert result.returncode == 0, f"Exited {result.returncode}: {' '.join(info_command)}"
        try:
            info_json = json.loads(str(result.stdout, "utf-8"))
        except json.JSONDecodeError:
            pytest.fail("Non JSON emitted from compiled connector")
        assert "response" in info_json, "Unexpected JSON structure from compiled connector"
        assert (
            "version" in info_json["response"]
        ), "Unexpected JSON structure from compiled connector"



import base64
import hashlib
import logging
import os
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any
from urllib.parse import quote_plus, urlencode

from connector_sdk_types.generated import (
    AuthorizationUrl,
    ErrorCode,
    GetAuthorizationUrlRequest,
    GetAuthorizationUrlResponse,
    HandleAuthorizationCallbackRequest,
    HandleAuthorizationCallbackResponse,
    HandleClientCredentialsRequest,
    HandleClientCredentialsResponse,
    OauthCredentials,
    RefreshAccessTokenRequest,
    RefreshAccessTokenResponse,
    StandardCapabilityName,
)
from httpx import BasicAuth

from connector.auth_helper import parse_auth_code_and_redirect_uri
from connector.httpx_rewrite import AsyncClient
from connector.oai.capability import AuthRequest
from connector.oai.errors import ConnectorError
from connector.oai.modules.base_module import BaseIntegrationModule
from connector.oai.modules.oauth_module_types import (
    ClientAuthenticationMethod,
    OAuthConfig,
    OAuthFlowType,
    OAuthRequest,
    OAuthSettings,
    RequestDataType,
    RequestMethod,
)

if TYPE_CHECKING:
    from connector.oai.integration import Integration

LOGGER = logging.getLogger("integration-connectors.sdk")


class OAuthModule(BaseIntegrationModule):
    """
    OAuth module is responsible for handling the OAuth2.0 authorization flow.
    It can register the following capabilities:
    - GET_AUTHORIZATION_URL
    - HANDLE_AUTHORIZATION_CALLBACK
    - REFRESH_ACCESS_TOKEN
    """

    def __init__(self, credentials: Sequence[OAuthSettings | OAuthConfig] | None = None):
        super().__init__()
        self.settings: list[OAuthSettings | OAuthConfig] = []

        if credentials:
            for cred in credentials:
                if isinstance(cred, OAuthSettings):
                    self.settings.append(cred)
                elif (
                    isinstance(cred, OAuthConfig)
                    and cred.oauth_settings is not None
                    and isinstance(cred.oauth_settings, OAuthSettings)
                ):
                    self.settings.append(cred)

    def register(self, integration: "Integration"):
        if len(self.settings) == 0:
            LOGGER.warning(
                f"OAuth settings were not provided for app '{integration.app_id}', skipping OAuth capabilities!"
            )
            return

        self.integration = integration

        # Pick up the enabled flows from either settings
        enabled_flows: set[OAuthFlowType] = set()
        for setting in self.settings:
            if isinstance(setting, OAuthConfig) and setting.oauth_settings is not None:
                enabled_flows.add(setting.oauth_settings.flow_type)
            elif isinstance(setting, OAuthSettings):
                enabled_flows.add(setting.flow_type)

        for flow_type in enabled_flows:
            # Default available capabilities for CODE FLOW
            if flow_type == OAuthFlowType.CODE_FLOW:
                capability_methods = {
                    StandardCapabilityName.GET_AUTHORIZATION_URL: self.register_get_authorization_url,
                    StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK: self.register_handle_authorization_callback,
                    StandardCapabilityName.REFRESH_ACCESS_TOKEN: self.register_refresh_access_token,
                }

            # Default available capabilities for CLIENT CREDENTIALS FLOW
            if flow_type == OAuthFlowType.CLIENT_CREDENTIALS:
                capability_methods = {
                    StandardCapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST: self.register_handle_client_credentials_request,
                    StandardCapabilityName.REFRESH_ACCESS_TOKEN: self.register_refresh_access_token,
                }

            # Register enabled capabilities
            for capability, register_method in capability_methods.items():
                # Prevent duplicate registration
                is_capability_registered = capability in self.capabilities
                is_capability_enabled = True

                for setting in self.settings:
                    if isinstance(setting, OAuthConfig) and setting.oauth_settings is not None:
                        is_capability_enabled = getattr(
                            setting.oauth_settings.capabilities, capability.value, True
                        )
                    elif isinstance(setting, OAuthSettings):
                        is_capability_enabled = getattr(
                            setting.capabilities, capability.value, True
                        )

                if not is_capability_registered and is_capability_enabled:
                    register_method()
                    self.add_capability(capability.value)

    def _get_settings(self, credential_id: str | None = None) -> OAuthSettings:
        if credential_id:
            # Find the settings for the given credential id
            # If not found, raise an error
            settings = next(
                (
                    setting
                    for setting in self.settings
                    if isinstance(setting, OAuthConfig) and setting.id == credential_id
                ),
                None,
            )
            if settings is None or settings.oauth_settings is None:
                raise ConnectorError(
                    message=f"OAuth settings for credential ID: {credential_id} were not provided!",
                    error_code=ErrorCode.BAD_REQUEST,
                )
            return settings.oauth_settings
        else:
            # If no credential id is provided, return the first settings object (legacy)
            if not self.settings[0] or not isinstance(self.settings[0], OAuthSettings):
                raise ConnectorError(
                    message=f"OAuth settings were not provided for app '{self.integration.app_id}'!",
                    error_code=ErrorCode.BAD_REQUEST,
                )
            return self.settings[0]

    def _get_url(self, url: str | Callable[[AuthRequest], str] | None, args: AuthRequest) -> str:
        if url is None:
            raise ConnectorError(
                message="Required URL was not provided for the OAuth flow.",
                error_code=ErrorCode.BAD_REQUEST,
            )
        if callable(url):
            return url(args)
        elif isinstance(url, str):
            return url

    def _get_scopes(
        self,
        settings: OAuthSettings,
        args: AuthRequest,
    ) -> str:
        """
        Get the scopes for the OAuth2.0 authorization flow from connector settings, formatted as a space delimited string.
        """
        scopes: dict[str, str] | Callable[[AuthRequest], dict[str, str]] | None = settings.scopes

        if scopes is None:
            # No scopes defined, return empty string
            return ""

        def parse_scopes(scope_dict: dict[str, str]) -> str:
            # May contain more than one value in the string for each scope
            string_scope_values = [
                value for value in scope_dict.values() if value is not None and value != ""
            ]
            # parse out multiple scopes
            scope_lists = [value.split(" ") for value in string_scope_values]
            # flatten and deduplicate
            scope_values = list(set(scope for sublist in scope_lists for scope in sublist))
            return " ".join(scope_values)

        if callable(scopes):
            # Call the connector defined method
            return parse_scopes(scopes(args))

        if isinstance(scopes, dict):
            # Return as space delimited string right away
            return parse_scopes(scopes)

    def _generate_code_challenge(self) -> tuple[str, str]:
        """
        Generate a code verifier and code challenge when using PKCE (Proof Key for Code Exchange).
        """
        code_verifier = base64.urlsafe_b64encode(os.urandom(32)).rstrip(b"=").decode("utf-8")
        code_challenge = (
            base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode("utf-8")).digest())
            .rstrip(b"=")
            .decode("utf-8")
        )
        return code_verifier, code_challenge

    async def _send_authorized_request(
        self,
        url: str,
        grant_type: str,
        client: AsyncClient,
        args: HandleAuthorizationCallbackRequest
        | RefreshAccessTokenRequest
        | HandleClientCredentialsRequest,
        credential_id: str | None = None,
    ) -> tuple[OauthCredentials, dict[str, Any]]:
        """
        Construct an authorized request to the token URL based on the grant type and request types.
        """
        settings = self._get_settings(credential_id)

        if grant_type == "authorization_code" and isinstance(
            args, HandleAuthorizationCallbackRequest
        ):
            # Handle authorization code request
            authorization_code, original_redirect_uri = parse_auth_code_and_redirect_uri(args)
            data = {
                "grant_type": grant_type,
                "code": authorization_code,
                "redirect_uri": original_redirect_uri,
            }

            # PKCE follow-up
            if args.request.code_verifier:
                data["code_verifier"] = args.request.code_verifier

        elif grant_type == "client_credentials" and isinstance(
            args, HandleClientCredentialsRequest
        ):
            # Handle client credentials request
            data = {
                "grant_type": grant_type,
            }

            # Some Client Credentials grant providers require the scope to be sent in the body/query
            scope = (
                " ".join(args.request.scopes)
                if args.request.scopes
                else self._get_scopes(settings, args)
            )
            if scope and scope != "":
                data["scope"] = scope

        elif grant_type == "refresh_token" and isinstance(args, RefreshAccessTokenRequest):
            # Handle refresh token request
            data = {
                "grant_type": grant_type,
                "refresh_token": args.request.refresh_token,
            }
        else:
            # Unsupported grant type
            raise ValueError(f"Unsupported grant_type: {grant_type}")

        # Some OAuth providers require client ID and secret to be sent in a Authorization header
        if settings.client_auth == ClientAuthenticationMethod.CLIENT_SECRET_BASIC:
            auth = BasicAuth(username=args.request.client_id, password=args.request.client_secret)
        else:
            # Others expect it in the body/query
            data.update(
                {
                    "client_id": args.request.client_id,
                    "client_secret": args.request.client_secret,
                }
            )
            auth = None

        # Default to POST and BODY if not specified in connector settings
        oauth_request_type = settings.request_type or OAuthRequest(
            method=RequestMethod.POST, data=RequestDataType.FORMDATA
        )
        request_method, request_data_type = oauth_request_type.method, oauth_request_type.data

        # Distribute data between query params and form-body/json
        if request_data_type == RequestDataType.QUERY:
            params = data
            body = None
            json = None
        elif request_data_type == RequestDataType.JSON:
            params = None
            body = None
            json = data
        else:
            params = None
            body = data
            json = None

        # Send the request
        response = await client.request(
            method=request_method,
            url=url,
            params=params,
            json=json,
            data=body,
            auth=auth,
        )

        # Raise for status
        if response.status_code >= 300:
            try:
                error_data = response.json()
                response_message = str(error_data)
            except ValueError:
                response_message = response.text
            response_code = response.status_code

            raise ConnectorError(
                message=f"[{response_code}][{url.split('?')[0]}] OAuth request failed: {response_message}",
                error_code=ErrorCode.API_ERROR,
            )

        # Convert token_type to lowercase if not specified
        response_json = response.json()
        response_json["token_type"] = (
            response_json["token_type"].lower() if "token_type" in response_json else "bearer"
        )

        oauth_credentials = OauthCredentials.from_dict(response_json)
        if oauth_credentials is None:
            raise ConnectorError(
                message="Unable to convert raw json to OauthCredentials",
                error_code=ErrorCode.BAD_REQUEST,
            )

        return oauth_credentials, response_json

    def register_get_authorization_url(self):
        @self.integration.register_capability(StandardCapabilityName.GET_AUTHORIZATION_URL)
        async def get_authorization_url(
            args: GetAuthorizationUrlRequest,
        ) -> GetAuthorizationUrlResponse:
            settings = self._get_settings(args.request.credential_id)

            url = self._get_url(settings.authorization_url, args)
            client_id = args.request.client_id
            redirect_uri = args.request.redirect_uri
            scope = (
                " ".join(args.request.scopes)
                if args.request.scopes
                else self._get_scopes(settings, args)
            )
            state = args.request.state

            params = {
                "client_id": client_id,
                "response_type": "code",
                "scope": scope,
                "redirect_uri": redirect_uri,
                "state": state,
            }

            authorization_url = f"{url}?{urlencode(params, quote_via=quote_plus)}"

            if settings.pkce:
                code_verifier, code_challenge = self._generate_code_challenge()
                authorization_url += f"&code_challenge={quote_plus(code_challenge)}"
                authorization_url += "&code_challenge_method=S256"
            else:
                code_verifier = None

            return GetAuthorizationUrlResponse(
                response=AuthorizationUrl(
                    authorization_url=authorization_url,
                    code_verifier=code_verifier,
                )
            )

        return get_authorization_url

    def register_handle_client_credentials_request(self):
        @self.integration.register_capability(
            StandardCapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST
        )
        async def handle_client_credentials_request(
            args: HandleClientCredentialsRequest,
        ) -> HandleClientCredentialsResponse:
            settings = self._get_settings(args.request.credential_id)

            async with AsyncClient() as client:
                url = self._get_url(settings.token_url, args)
                oauth_credentials, response_json = await self._send_authorized_request(
                    url, "client_credentials", client, args
                )

                return HandleClientCredentialsResponse(
                    response=oauth_credentials,
                    raw_data=response_json if args.include_raw_data else None,
                )

        return handle_client_credentials_request

    def register_handle_authorization_callback(self):
        @self.integration.register_capability(StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK)
        async def handle_authorization_callback(
            args: HandleAuthorizationCallbackRequest,
        ) -> HandleAuthorizationCallbackResponse:
            settings = self._get_settings(args.request.credential_id)

            async with AsyncClient() as client:
                url = self._get_url(settings.token_url, args)
                oauth_credentials, response_json = await self._send_authorized_request(
                    url, "authorization_code", client, args
                )

                return HandleAuthorizationCallbackResponse(
                    response=oauth_credentials,
                    raw_data=response_json if args.include_raw_data else None,
                )

        return handle_authorization_callback

    def register_refresh_access_token(self):
        @self.integration.register_capability(StandardCapabilityName.REFRESH_ACCESS_TOKEN)
        async def refresh_access_token(
            args: RefreshAccessTokenRequest,
        ) -> RefreshAccessTokenResponse:
            settings = self._get_settings(args.request.credential_id)

            async with AsyncClient() as client:
                url = self._get_url(settings.token_url, args)
                oauth_credentials, response_json = await self._send_authorized_request(
                    url, "refresh_token", client, args
                )

                return RefreshAccessTokenResponse(
                    response=oauth_credentials,
                    raw_data=response_json if args.include_raw_data else None,
                )

        return refresh_access_token



"""Tests for ``connector.oai.integration`` module."""

import io
import sys

import pytest_cases
from connector.oai.integration import Integration
from connector.oai.modules.base_module import BaseIntegrationModule
from connector_sdk_types.generated import StandardCapabilityName

from tests.oai.test_modules_cases import (
    case_add_module,
    case_register_module_capabilities,
)


@pytest_cases.parametrize_with_cases(
    ["integration", "module"],
    cases=[
        case_add_module,
    ],
)
async def test_add_module(
    integration: Integration,
    module: BaseIntegrationModule,
) -> None:
    captured_output = io.StringIO()
    sys.stdout = captured_output

    try:
        integration.add_module(module)

        assert captured_output.getvalue().strip() == "registered module"
        assert module in integration.modules
    finally:
        sys.stdout = sys.__stdout__


@pytest_cases.parametrize_with_cases(
    ["integration", "module"],
    cases=[
        case_register_module_capabilities,
    ],
)
async def test_register_module_capabilities(
    integration: Integration,
    module: BaseIntegrationModule,
) -> None:
    integration.add_module(module)

    assert integration.capabilities[StandardCapabilityName.VALIDATE_CREDENTIALS] is not None
    assert integration.info().response.capabilities == [
        StandardCapabilityName.APP_INFO,
        StandardCapabilityName.VALIDATE_CREDENTIALS,
    ]
    assert module.capabilities == [StandardCapabilityName.VALIDATE_CREDENTIALS]



"""Cases for testing ``list_accounts`` capability."""

import typing as t

import httpx
from connector.generated import (
    ErrorResponse,
    ListAccounts,
    ListAccountsRequest,
    ListAccountsResponse,
    StandardCapabilityName,
)
from connector.tests.type_definitions import MockedResponse, ResponseBodyMap
from {name}.pagination import NextPageToken, Pagination

from tests.common_mock_data import SETTINGS, TEST_MAX_PAGE_SIZE, VALID_AUTH

Case: t.TypeAlias = tuple[
    StandardCapabilityName,
    ListAccountsRequest,
    ResponseBodyMap,
    ListAccountsResponse | ErrorResponse,
]


def case_list_accounts_200() -> Case:
    """Successful request."""
    args = ListAccountsRequest(
        request=ListAccounts(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=5&offset=0": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListAccountsResponse(
        response=[],
        page=NextPageToken.from_paginations(
            [Pagination(endpoint="", offset=2)]).to_page(size=TEST_MAX_PAGE_SIZE)
    )
    return StandardCapabilityName.LIST_ACCOUNTS, args, response_body_map, expected_response


def case_list_accounts_200_no_accounts() -> Case:
    """No accounts found."""
    args = ListAccountsRequest(
        request=ListAccounts(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=5&offset=0": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListAccountsResponse(
        response=[],
    )
    return StandardCapabilityName.LIST_ACCOUNTS, args, response_body_map, expected_response



from connector.generated import AuthCredential, OAuthCredential
from {name}.settings import {pascal}Settings

VALID_AUTH = AuthCredential(
    oauth=OAuthCredential(access_token="valid"),  # noqa: S105
)
INVALID_AUTH = AuthCredential(
    oauth=OAuthCredential(access_token="invalid"),  # noqa: S105
)
SETTINGS = {pascal}Settings(account_id="test-account-id").model_dump()
TEST_MAX_PAGE_SIZE = 100



import asyncio
import logging
from dataclasses import dataclass
from typing import Any

from connector.config import config
from connector.httpx_rewrite import AsyncClient


@dataclass
class LogInput:
    agent_identifier: str
    log: str

    def model_dump(self) -> dict[str, Any]:
        return {
            "agent_identifier": self.agent_identifier,
            "log": self.log,
        }


# Create a separate logger for this module that only uses stdout
# This is to avoid recursive calls to the logger
http_logger = logging.getLogger("lumos_log_handler")
# Prevent this logger from inheriting the custom https handler which would call itself
http_logger.propagate = False
# Add a basic stdout handler
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter("%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s"))
http_logger.addHandler(handler)


class LumosLogHandler(logging.Handler):
    def __init__(self) -> None:
        super().__init__()
        if not config.agent_identifier:
            http_logger.error("agent_identifier must be set in config")
            raise ValueError("agent_identifier must be set in config")
        self.agent_identifier = config.agent_identifier

    def emit(self, record: logging.LogRecord) -> None:
        # Only process ERROR level logs for now
        if record.levelno < logging.ERROR:
            return

        try:
            log_data = LogInput(
                agent_identifier=self.agent_identifier,
                log=self.format(record),
            )

            # Create event loop if one doesn't exist
            try:
                loop = asyncio.get_event_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)

            # Run the async send operation
            loop.run_until_complete(self._send_log(log_data))
        except Exception:
            http_logger.error("Failed to emit log record", exc_info=True)
            self.handleError(record)

    async def _send_log(self, log_data: LogInput) -> None:
        if not config.base_url:
            http_logger.error("base_url must be set in config")
            return
        if not config.on_prem_agent_api_key:
            http_logger.error("on_prem_agent_api_key must be set in config")
            return

        try:
            headers = {
                "Authorization": f"Bearer {config.on_prem_agent_api_key}",
                "X-Agent-Identifier": self.agent_identifier,
                "Content-Type": "application/json",
            }

            async with AsyncClient(
                base_url=config.base_url,
                headers=headers,
            ) as client:
                try:
                    response = await client.post(
                        "/logs",
                        json=log_data.model_dump(),
                    )
                    response.raise_for_status()
                except Exception as e:
                    http_logger.error(f"Failed to send log to Lumos: {e}", exc_info=True)
        except Exception as e:
            http_logger.error(f"Failed to send log to Lumos: {e}", exc_info=True)



from connector.utils.account import split_name


def test_split_name():
    assert split_name("John Doe") == ("John", "Doe")
    assert split_name("John") == ("John", None)
    assert split_name(None) == (None, None)
    assert split_name("John Doe Smith") == ("John", "Doe Smith")



import asyncio
import typing as t
from collections import abc
from functools import partial, wraps

RET = t.TypeVar("RET")
P = t.ParamSpec("P")
SelfType = t.TypeVar("SelfType")


def sync_to_async(
    func: abc.Callable[t.Concatenate[SelfType, P], RET],
) -> abc.Callable[t.Concatenate[SelfType, P], abc.Awaitable[RET]]:
    """
    Utility decorator to convert a synchronous method into an asynchronous one.
    """

    @wraps(func)
    async def wrapper(self: SelfType, *args: P.args, **kwargs: P.kwargs) -> RET:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, partial(func, self, *args, **kwargs))

    return wrapper



import typing as t

from graphql import DocumentNode, print_ast
from pytest_httpx import HTTPXMock

from connector.tests.type_definitions import ResponseBodyMap


def mock_graphql_request_body(
    query: DocumentNode, variables: dict[str, t.Any] | None = None
) -> dict[str, t.Any]:
    body: dict[str, t.Any] = {"query": print_ast(query)}

    if variables:
        body.update({"variables": variables})

    return body


def mock_requests(
    response_body_map: ResponseBodyMap, httpx_mock: HTTPXMock, *, host: str | None = None
):
    if not response_body_map:
        # Don't mock any requests, and use the default behavior of
        # httpx_mock which is HTTP 200, empty body
        return

    for method, responses in response_body_map.items():
        if isinstance(responses, dict):
            for request_line, response in responses.items():
                if request_line.startswith("https://"):
                    url = request_line
                else:
                    url = f"{host or ''}{request_line}"

                httpx_mock.add_response(
                    method=method,
                    url=url,
                    json=response.response_body,
                    text=(
                        response.response_body if isinstance(response.response_body, str) else None
                    ),
                    status_code=response.status_code,
                    headers=response.headers if response.headers else None,
                    match_json=response.request_json_body,
                    match_content=response.request_bytes_body,
                )
        elif isinstance(responses, list):
            for response in responses:
                httpx_mock.add_response(
                    method=method,
                    url=(
                        f"{host or ''}{response.request_path}"
                        if response.request_path
                        else host or ""
                    ),
                    json=response.response_body,
                    text=(
                        response.response_body if isinstance(response.response_body, str) else None
                    ),
                    status_code=response.status_code,
                    headers=response.headers if response.headers else None,
                    match_json=response.request_json_body,
                    match_content=response.request_bytes_body,
                )



import msgpack
import pydantic
import pytest
import pytest_cases

from tests.utils.definitions import NextPageToken, Pagination


async def test_encode_empty_paginations() -> None:
    """Test if empty paginations are encoded to ``None`` token."""
    token = NextPageToken.from_paginations([]).token
    assert token is None


@pytest_cases.parametrize_with_cases(
    ["paginations"],
    cases=["tests.utils.test_pagination_encode_cases"],
)
async def test_encode_pagination(
    paginations: list[Pagination],
) -> None:
    """Test if token is created from valid input."""
    token = NextPageToken.from_paginations(paginations).token
    assert isinstance(token, str)


async def test_decode_none_token() -> None:
    """Test if ``None`` token gives empty paginations."""
    paginations = NextPageToken(None).paginations()
    assert paginations == []


@pytest_cases.parametrize_with_cases(
    ["token"],
    cases=["tests.utils.test_pagination_decode_cases"],
)
async def test_decode_token(
    token: str,
) -> None:
    """Test if token is correctly decoded from valid input."""
    paginations = NextPageToken(token).paginations()
    assert isinstance(paginations, list)


@pytest_cases.parametrize_with_cases(
    ["paginations"],
    cases=["tests.utils.test_pagination_duality_cases"],
)
async def test_decode_encode_pagination(
    paginations: list[Pagination],
) -> None:
    """Test duality of ``from_paginations`` and ``paginations``.

    This test simply checks if the input paginations can correctly be
    decoded.
    """
    decoded_paginations = NextPageToken.from_paginations(paginations).paginations()
    assert decoded_paginations == paginations


async def test_invalid_byte_token() -> None:
    """Test if invalid bytes are rejected as pagination token."""
    token = "1234567890abcdef"
    with pytest.raises((msgpack.exceptions.UnpackException, msgpack.exceptions.ExtraData)):
        NextPageToken(token).paginations()


async def test_valid_token_invalid_data() -> None:
    """Test if data that do not match pagination schema are rejected.

    The token contains incorrect key for ``page``, due to typo there is
    ``pgae``.
    """
    invalid_pagination = [
        {
            "endpoint": "/users",
            "pgae": 3,  # tpyo is here
        },
    ]
    token = msgpack.packb(invalid_pagination).hex()
    with pytest.raises(pydantic.ValidationError) as excinfo:
        NextPageToken(token).paginations()

    invalid_fields_with_error_types = [
        (error["loc"], error["type"]) for error in excinfo.value.errors()
    ]
    expected_errors = [
        (("page",), "missing"),
        (("pgae",), "extra_forbidden"),
    ]
    assert set(invalid_fields_with_error_types) == set(expected_errors)



import os
import pathlib
import re
import sys

from connector.oai.capability import StandardCapabilityName
from connector.oai.integration import Integration


def gather_cases(
    integration: Integration,
    capabilities_to_test: list[str] | None = None,
    tests_dir_path: pathlib.Path | None = None,
) -> list[str]:
    """
    Gather test cases for the given integration. Tries to discover test cases based on location of integration instance.
    Goes through all test files and checks if they are named within convention: test_{capability_name}_cases.py.
    Validates if all capabilities have corresponding test cases, if not, adds a coverage check.
    """
    if tests_dir_path is None:
        if not isinstance(integration.app_id, str):
            raise ValueError(
                "Integration app_id must be a string, please check your test_all_cases.py and consider using static paths in args to gather_cases."
            )

        module_path: str | None = None
        try:
            module = sys.modules[integration.app_id]
            if module and hasattr(module, "__file__"):
                module_path = module.__file__
        except Exception:
            """
            In case of no module access, try to get the folder of the caller via inspect
            This is used for example when you want to detach a single module and run
            it in a Docker container, without any of the other modules or actual module access.
            """
            import inspect

            caller_frame = inspect.currentframe()
            if caller_frame and caller_frame.f_back:
                caller_path = inspect.getfile(caller_frame.f_back)
                caller_folder = pathlib.Path(caller_path).resolve()
                if caller_folder:
                    module_path = str(caller_folder)

        if module_path:
            integration_path = pathlib.Path(module_path).parent.parent
            tests_dir_path = integration_path / "tests"
        else:
            raise ValueError(
                f"Could not find module path for {integration.app_id}, please check your test_all_cases.py and consider using static paths in args to gather_cases."
            )

    if capabilities_to_test is None:
        capabilities_to_test = [value for value in integration.capabilities.keys()]

    files = []
    for capability in capabilities_to_test:
        pattern = rf"test_(?:.+_)*?{capability}(?:_.+)*_cases.py"
        compiled = re.compile(pattern)
        matching_files = [
            file for file in tests_dir_path.rglob("*.py") if compiled.fullmatch(file.name)
        ]

        for file_path in matching_files:
            relative_path = file_path.relative_to(tests_dir_path)
            module_path = str(relative_path.with_suffix("")).replace(os.sep, ".")
            files.append(f"tests.{module_path}")

    has_full_coverage, missing_capabilities = calculate_coverage(integration, files)
    if not has_full_coverage:
        import connector.tests.coverage_check as coverage_check

        coverage_check.missing_capabilities = missing_capabilities
        coverage_check.integration_name = integration.app_id
        files.append("connector.tests.coverage_check")

    return files


def calculate_coverage(integration: Integration, cases: list[str]) -> tuple[bool, set[str]]:
    """
    Checks if all registered capabilities have corresponding test case files.
    Ignores system wide capabilities that have their own module and tests.
    """
    ignored_capabilities = [
        StandardCapabilityName.HANDLE_AUTHORIZATION_CALLBACK,
        StandardCapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST,
        StandardCapabilityName.REFRESH_ACCESS_TOKEN,
        StandardCapabilityName.GET_AUTHORIZATION_URL,
        StandardCapabilityName.APP_INFO,
    ]
    capabilities_to_test = [
        value
        for value in integration.capabilities.keys()
        if value not in ignored_capabilities and value in [c.value for c in StandardCapabilityName]
    ]
    existing_capabilities = set(
        capability
        for file in cases
        for capability in capabilities_to_test
        if re.search(f"test_(?:.+_)*?{capability}(?:_.+)*_cases", file) is not None
    )
    required_capabilities = set(capability for capability in capabilities_to_test)
    missing_capabilities = required_capabilities - existing_capabilities

    coverage = len(existing_capabilities) / len(required_capabilities)
    return coverage >= 1, missing_capabilities



import contextvars
import logging
import re
from collections.abc import Generator
from contextlib import contextmanager

import httpx
from connector_sdk_types.generated import ErrorCode
from gql.transport.httpx import HTTPXAsyncTransport as GqlHTTPXAsyncTransport
from urllib3.util.url import Url, parse_url

from connector.oai.errors import ConnectorError

logger = logging.getLogger("integration-connectors.sdk")
GLOBAL_REQUEST_TIMEOUT = httpx.Timeout(300.0)  # 5 minutes


class ResponseLogRecord(logging.LogRecord):
    method: str
    url: str
    status_code: int
    headers: dict[str, str]
    content: str


class ResponseLogFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record)

        if isinstance(record, ResponseLogRecord):
            extras = {
                "method": record.method,
                "url": record.url,
                "status_code": record.status_code,
                "headers": record.headers,
                "content": record.content[:1000] + "..."
                if len(record.content) > 1000
                else record.content,
            }
            message = f"{message}\n Details: {extras}"

        return message


response_logger = logging.getLogger("integration-connectors.sdk.api_response")
handler = logging.StreamHandler()
handler.setFormatter(ResponseLogFormatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
response_logger.addHandler(handler)

LUMOS_PROXY_URL: contextvars.ContextVar[str | None] = contextvars.ContextVar(
    "LUMOS_PROXY_URL", default=None
)
LUMOS_PROXY_HEADERS: contextvars.ContextVar[httpx.Headers] = contextvars.ContextVar(
    "LUMOS_PROXY_HEADERS", default=httpx.Headers()
)

DENYLISTED_HEADERS = {
    "authorization",
    "auth",
    "token",
    "api-key",
    "apikey",
    "x-api-key",
    "client-secret",
    "client_secret",
    "bearer",
    "jwt",
    "session",
    "cookie",
    "set-cookie",
    "x-auth",
    "x-auth-token",
    "basic",
    "password",
    "secret",
    "private-key",
    "access-key",
    "access_key",
}
REDACTION_VALUE = "[REDACTED]"
DATADOG_TRUNCATION_LIMIT = 24 * 1024 * 1024
CONTENT_MAX_LENGTH = DATADOG_TRUNCATION_LIMIT - 64
SENSITIVE_FIELDS_PATTERN = (
    r"(access_token|refresh_token|temporary_password|token|api[_-]?key|client[_-]?secret|"
    r"password|secret|auth[_-]?token|jwt|bearer|"
    r"ssn|social[_-]?security|tax[_-]?id|ein|"
    r"national[_-]?id|passport[_-]?number|driver[_-]?license|"
    r"date[_-]?of[_-]?birth|birth[_-]?date|dob|"
    r"phone|mobile|cell|telephone|"
    r"email|mail|"
    r"address[_-]?line[0-9]|street|city|state|zip|postal|country|"
    r"card[_-]?number|cvv|cvc|pin|account[_-]?number)"
)


@contextmanager
def proxy_settings(
    *, proxy_url: str | None, proxy_headers: dict[str, str] | httpx.Headers | None = None
) -> Generator[None, None, None]:
    """
    Set context to run some httpx requests that are rewritten to a proxy.

    Example:

      from httpx_rewrite import AsyncClient, proxy_settings

      with proxy_settings(proxy_url="https://mitm.lumos.com/whatever/path", proxy_headers={"password":"swordfish"}):
          client = AsyncClient()
          await client.get("https://a-real-url-to-call/different/path")
    """
    LUMOS_PROXY_URL.set(proxy_url)
    LUMOS_PROXY_HEADERS.set(httpx.Headers(proxy_headers))
    yield


def get_proxy_url() -> str | None:
    """Return the current URL requests should proxy to. Set this with proxy_settings()"""
    return LUMOS_PROXY_URL.get()


def get_proxy_headers() -> httpx.Headers:
    """Return the current headers that should be attached to proxied requests. Set this with proxy_settings()"""
    return LUMOS_PROXY_HEADERS.get()


def log_response(response: httpx.Response):
    content = response.text

    if len(content) > CONTENT_MAX_LENGTH:
        content = content[: CONTENT_MAX_LENGTH // 2] + " ... " + content[-CONTENT_MAX_LENGTH // 2 :]
    content = re.sub(
        f'"{SENSITIVE_FIELDS_PATTERN}":\\s*"[^"]*"',
        f'"\\1": "{REDACTION_VALUE}"',
        content,
        flags=re.IGNORECASE,
    )

    url = str(response.url)
    url = re.sub(
        f"[?&]({SENSITIVE_FIELDS_PATTERN})=[^&]*",
        f"\\1={REDACTION_VALUE}",
        url,
        flags=re.IGNORECASE,
    )

    headers = {}
    for key, value in response.headers.items():
        if any(denylisted_header in key.lower() for denylisted_header in DENYLISTED_HEADERS):
            headers[key] = REDACTION_VALUE
        else:
            headers[key] = value

    content_hash = ""
    try:
        content_hash = str(hash(content))
    except Exception:
        content_hash = "hashing failed"

    record = ResponseLogRecord(
        name=response_logger.name,
        level=response_logger.level,
        pathname="",
        lineno=0,
        msg=f"Got an external response: [{response.request.method}][{response.status_code}] {url}",
        args=(),
        exc_info=None,
    )

    record.method = str(response.request.method)
    record.url = url
    record.status_code = response.status_code
    record.headers = headers
    record.content = content
    record.content_hash = content_hash

    response_logger.handle(record)


class AsyncClient(httpx.AsyncClient):
    """
    A slightly modified version of httpx.AsyncClient, that may rewrite requests to go to a proxy.

    Requests will be rewritten when run within the proxy_settings() context.

    Rewritten requests...
    1. will go to the proxy URL
    2. will have the original request line (including fragment) preserved in the header `X-Forward-To`
    3. will have their headers preserved, EXCEPT FOR...
       a. `Host` will be set to the proxy host
       b. Additional headers will be set, and overwrite any passed-in headers, from
         - the LUMOS_PROXY_HEADERS environment variable (as JSON)
         - the accumulated calls to this module's `add_proxy_headers` function
    """

    def will_rewrite_request(self) -> bool:
        return bool(get_proxy_url())

    def __init__(self, *args, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = GLOBAL_REQUEST_TIMEOUT
        super().__init__(*args, **kwargs)

    async def send(self, request: httpx.Request, **kwargs) -> httpx.Response:
        proxy_host = get_proxy_url()
        if proxy_host:
            logger.warning("Proxying all requests to %s", proxy_host)
            override = parse_url(proxy_host)
        else:
            response = await super().send(request, **kwargs)
            log_response(response)
            return response

        # Update request URL
        url = str(request.url)
        updated = Url(
            scheme=override.scheme,
            host=override.host,
            port=override.port,
            path=override.path,
            query=override.query,
        )

        # Rewrite headers
        new_headers = request.headers.copy()
        new_headers["X-Forward-To"] = url
        new_headers["Host"] = Url(host=override.host, port=override.port).url
        proxy_headers = get_proxy_headers()
        for header, value in proxy_headers.items():
            new_headers[header] = value

        # Get request body if applicable
        if request.method in ("POST", "PUT", "PATCH"):
            content = request.content or bytes([])
        else:
            content = None

        new_request = httpx.Request(
            method=request.method,
            url=updated.url,
            headers=new_headers,
            content=content,
        )

        # Log everything when using the proxy for debugging
        logger.debug("Sending request to proxy: %s", new_request.url)
        logger.debug("Request method: %s", new_request.method)
        response = await super().send(new_request, **kwargs)
        logger.debug("Request headers: %s", [(k, v) for k, v in response.request.headers.items()])
        logger.debug("Request body: %s", response.request.content)
        logger.debug("Response headers: %s", [(k, v) for k, v in response.headers.items()])
        truncated_response = response.text[:1000] + "..." if len(response.text) > 1000 else ""
        logger.debug("Response: %s", truncated_response)
        logger.debug("Response status code: %s", response.status_code)

        # Create a debug hint if the proxy/service is not working as expected
        proxy_status = int(response.headers.get("x-lumos-sip-status-code") or response.status_code)
        target_status = response.headers.get("x-lumos-target-hostname-status-code")
        if proxy_status != 200:
            hint = (
                "There is something wrong with the proxy itself or you are giving it a bad "
                f"arn value. Proxy returned code {proxy_status}; "
                "sort this out first before troubleshooting the target service."
            )
            hint_status = proxy_status
            hint_url = new_request.url

        elif target_status is not None and not str(target_status).startswith("2"):
            hint = (
                f"The target service returned code {target_status}, "
                "the proxy is working as expected."
            )
            hint_status = target_status
            hint_url = request.url

        if "hint" in locals():
            raise ConnectorError(
                error_code=ErrorCode.INVALID_RESPONSE,
                message=f"[{hint_status}][{hint_url}] {hint} Original response: {response.text}",
                app_error_code="proxy_mode_error_debug_hint",
            )

        return response


class HTTPXAsyncTransport(GqlHTTPXAsyncTransport):
    async def connect(self):
        # Call the superclass...
        await super().connect()

        # ... and then override its result
        self.client = AsyncClient(**self.kwargs)



from connector.generated import (
    ActivateAccountRequest,
    ActivateAccountResponse,
    AssignEntitlementRequest,
    AssignEntitlementResponse,
    CreateAccountResponse,
    DeactivateAccountRequest,
    DeactivateAccountResponse,
    DowngradeLicenseRequest,
    DowngradeLicenseResponse,
    DeleteAccountRequest,
    DeleteAccountResponse,
    ReleaseResourcesRequest,
    ReleaseResourcesResponse,
    TransferDataRequest,
    TransferDataResponse,
    UnassignEntitlementRequest,
    UnassignEntitlementResponse,
)
from connector.integration import CustomRequest

from {name}.dto.user import CreateAccount


async def assign_entitlement(args: AssignEntitlementRequest) -> AssignEntitlementResponse:
    raise NotImplementedError


async def unassign_entitlement(
    args: UnassignEntitlementRequest,
) -> UnassignEntitlementResponse:
    raise NotImplementedError


async def create_account(
    args: CustomRequest[CreateAccount],
) -> CreateAccountResponse:
    raise NotImplementedError


async def delete_account(
    args: DeleteAccountRequest,
) -> DeleteAccountResponse:
    raise NotImplementedError


async def activate_account(
    args: ActivateAccountRequest,
) -> ActivateAccountResponse:
    raise NotImplementedError


async def deactivate_account(
    args: DeactivateAccountRequest,
) -> DeactivateAccountResponse:
    raise NotImplementedError

async def transfer_data(
    args: TransferDataRequest,
) -> TransferDataResponse:
    raise NotImplementedError

async def downgrade_license(
    args: DowngradeLicenseRequest,
) -> DowngradeLicenseResponse:
    raise NotImplementedError

async def release_resources(
    args: ReleaseResourcesRequest,
) -> ReleaseResourcesResponse:
    raise NotImplementedError



import asyncio
import copy
import gzip
import json
import logging
import os
import subprocess
import sys
from argparse import ArgumentParser, Namespace, RawDescriptionHelpFormatter
from datetime import datetime, timezone
from typing import Any

from connector.ca_certs import is_windows, set_python_to_use_system_ca_certificates
from connector.config import config
from connector.httpx_rewrite import proxy_settings
from connector.logging import set_logger_config
from connector.oai.integration import Integration
from connector.pydantic import get_pydantic_model
from connector.utils import proxy_utils

logger = logging.getLogger("integration-connectors.sdk")

# Hacking commands
# ----------------


def _prep_hacking_command(args: Namespace):
    data = vars(args)
    data.pop("command")
    data.pop("func")
    return data


def http_integration_server(
    integration: Integration, port: int = 8000, reload: bool = False, use_proxy: bool = False
):
    from connector.http_server import runserver

    try:
        runserver(port=port, reload=reload, integration=integration, use_proxy=use_proxy)
    except KeyboardInterrupt:
        pass


def build_executable(path: str) -> None:
    try:
        subprocess.run(["pyinstaller", "--version"], check=True)
    except FileNotFoundError:
        print("PyInstaller not found in PATH. Please pip install pyinstaller")
        return

    command = [
        "pyinstaller",
        path,
        "--noconsole",
        "--onefile",
        "--clean",
        "--paths=projects/libs/python/connector-sdk",
    ]
    if __file__ not in "site-packages":
        command.append("--paths=projects/libs/python/connector-sdk")
    subprocess.run(command)


def create_integration_hacking_parser(integration: Integration, parser: ArgumentParser) -> None:
    subparsers = parser.add_subparsers(dest="command")

    http_server_parser = subparsers.add_parser(
        "http-server",
        help="Run this connector as an HTTP server.",
        description="\n".join(
            [
                "Run this connector as an HTTP server.",
                " - You can call capabilities via POST /<capability name> with the input JSON as the request body.",
                " - API docs are at /docs",
                " - The OpenAPI spec is at /openapi.json",
            ]
        ),
        formatter_class=RawDescriptionHelpFormatter,
    )
    http_server_parser.add_argument(
        "--port", "-p", type=int, default=8000, help="The port to run the server on."
    )
    http_server_parser.add_argument(
        "--reload", action="store_true", help="Enable hot reload for the server."
    )
    http_server_parser.add_argument(
        "--use-proxy", "--sip", action="store_true", help="Use the Lumos SIP to proxy requests."
    )
    http_server_parser.set_defaults(
        func=lambda args: http_integration_server(integration, **_prep_hacking_command(args))
    )

    build_executable_parser = subparsers.add_parser(
        "build-executable",
        help=(
            "Create a single file executable with PyInstaller. Provide the path to your library's"
            " main.py file."
        ),
    )
    build_executable_parser.add_argument("path", type=str, help="The path to the main.py file.")
    build_executable_parser.set_defaults(
        func=lambda args: build_executable(**_prep_hacking_command(args))
    )

    return None


# Actual Commands
# ---------------


def get_result_file_path(args: Namespace) -> str | None:
    if not args.result_file_path:
        return None
    result_file_path = args.result_file_path.strip('"').strip("'")
    if not result_file_path.endswith(".gz"):
        raise ValueError("The result file name must end with the .gz extension.")
    return result_file_path


def get_json_value(args: Namespace) -> str | None:
    if not args.json and not args.json_file:
        return None
    if args.json_file:
        with open(args.json_file.strip('"').strip("'")) as f:
            return f.read()
    return args.json


def capability_executor(integration: Integration, args: Namespace):
    """Executes a command from the CLI."""
    # validate that a valid gzip file name was provided
    result_file_path = get_result_file_path(args)

    if args.command == "info":
        output = json.dumps(integration.info().model_dump(), sort_keys=True)
    else:
        json_value = get_json_value(args)
        if json_value is None:
            raise ValueError("No JSON value provided")
        if args.use_proxy:
            with proxy_settings(
                proxy_url=proxy_utils.get_proxy_url(),
                proxy_headers={
                    "X-Lumos-Proxy-Auth": proxy_utils.get_proxy_token_sync().token,
                },
            ):
                output = asyncio.run(integration.dispatch(args.command, json_value))
        else:
            output = asyncio.run(integration.dispatch(args.command, json_value))

    if result_file_path:
        logger.info(f"Attempting to open file name: {result_file_path}")
        with open(result_file_path, "w") as result_file:
            logger.info(f"File opened to write: {result_file_path}")
            logger.info("compressing result")
            output_bytes = output.encode("utf-8")
            gzipped_result = gzip.compress(output_bytes)
            logger.info("writing compressed result to file")
            result_file.buffer.write(gzipped_result)
            logger.info("compressed result written to file")
        logger.info(f"Result saved to {args.result_file_path}")
    else:
        logger.info("Result printing to console")
        print(output)

    logger.info("Command completed")


CAPABILITY_PREFIX = "\n      "


def collect_capabilities(integration: Integration, no_print: bool = False) -> ArgumentParser:
    """
    Collect all methods from an Integration class and create a CLI
    command for each.
    """
    executed = os.path.basename(sys.argv[0])
    capability_helps: list[str] = []
    capability_helps = sorted(integration.capabilities.keys())
    parser = ArgumentParser(
        description=f"Lumos integration CLI for {integration.description_data.user_friendly_name}",
        usage=f"""{executed} CAPABILITY [--json JSON STRING] [--json-file JSON FILEPATH] [--result-file-path FILEPATH]

    Examples:

    {executed} info
        Print the Info schema for how to call this connector, and what it supports. This is the only
        capability that takes no arguments.

    {executed} validate_credentials --json '{"{}"}'
        Check if you're passing enough auth credentials and settings to connect to the underlying
        app tenant.

    All capabilities except 'info' require a JSON argument or a JSON file argument.

    A typical JSON argument looks like

    {"{"}
        "auth": {"{...}"},
        "settings": {"{...}"},
        "request": {"{...}"}
    {"}"}

    All capabilities:

{CAPABILITY_PREFIX}{CAPABILITY_PREFIX.join([c for c in capability_helps])}

""",
    )

    subparsers = parser.add_subparsers(dest="command")

    subparser = subparsers.add_parser("info", description=integration.info.__doc__)
    _add_arguments_to_subparser(subparser)
    subparser.set_defaults(func=lambda args: capability_executor(integration, args))

    for capability_name, capability in integration.capabilities.items():
        subparser = subparsers.add_parser(capability_name, description=capability.__doc__)

        try:
            get_pydantic_model(capability.__annotations__)
        except ValueError:
            pass
        else:
            group = subparser.add_mutually_exclusive_group(required=True)
            group.add_argument("--json", type=str, help="JSON input")
            group.add_argument("--json-file", type=str, help="JSON file input")
            _add_arguments_to_subparser(subparser)

        subparser.set_defaults(func=lambda args: capability_executor(integration, args))

    hacking_subparser = subparsers.add_parser("dev", aliases=["hacking"])
    create_integration_hacking_parser(integration, hacking_subparser)

    return parser


def _add_arguments_to_subparser(subparser: ArgumentParser):
    subparser.add_argument(
        "--result-file-path",
        type=str,
        help=(
            "The path to the file to save the result to. If this file already exists, "
            "its content will be overwritten. If not provided, the result will be "
            "printed to the console. The result file will be gzip compressed so the "
            "input file name should end with the .gz extension."
        ),
        default=None,
    )
    subparser.add_argument(
        "--use-proxy",
        "--sip",
        action="store_true",
        help="Use the Lumos SIP to proxy requests. This should not be used by non-Lumos developers.",
    )


def run_integration(
    integration: Integration,
    no_print: bool = False,
) -> None:
    set_logger_config(integration.app_id)
    if is_windows():
        set_python_to_use_system_ca_certificates()
    logger.info("Running command started at %s", datetime.now(timezone.utc))
    try:
        """Run a command from the CLI, integration version."""
        parser = collect_capabilities(integration, no_print)
        args = parser.parse_args()
        try:
            logger.info("Command arguments: %s", build_loggable_args(args, integration))
        except Exception:
            logger.exception("Error building loggable arguments")
        if not args.command:
            print("No command passed in", file=sys.stderr)
            parser.print_help(file=sys.stderr)
            sys.exit(1)
        args.func(args)
    except Exception as e:
        logger.error(
            f"Error running command exception class: {e.__class__.__name__} exception: {e}"
        )
        logger.error("Stack trace:", exc_info=True)
        raise e
    logger.info("Command completed at %s", datetime.now(timezone.utc))


def build_loggable_args(args: Namespace, integration: Integration) -> dict[str, Any]:
    loggable_args = copy.deepcopy(args.__dict__)
    if loggable_args.get("json"):
        json_obj = json.loads(args.json)
        redact_json_obj(json_obj, get_integration_secret_fields(integration))
        loggable_args["json"] = json_obj
    return loggable_args


BASE_REDACTED_LOG_KEYS = [
    "password",
    "secret",
    "token",
    "api_key",
    "api_token",
    "api_secret",
    "access_token",
    "refresh_token",
    "secret",
    "client_secret",
    "consumer_secret",
    "token_secret",
]


def get_integration_secret_fields(integration: Integration) -> list[str]:
    all_secret_fields = []
    secret_settings_fields = [
        name
        for name, field_info in integration.settings_model.model_fields.items()
        # Mypy doesn't recognize the None check here
        if field_info.json_schema_extra is not None and field_info.json_schema_extra.get("x-secret")  # type: ignore
    ]
    all_secret_fields.extend(secret_settings_fields)

    integration_auth_model = integration.auth
    if integration_auth_model is not None:
        secret_auth_fields = [
            name
            for name, field_info in integration_auth_model.model_fields.items()
            if field_info.json_schema_extra is not None
            and field_info.json_schema_extra.get("x-secret")  # type: ignore
        ]
        all_secret_fields.extend(secret_auth_fields)
    return all_secret_fields


def redact_json_obj(json_obj: Any, secret_fields: list[str]) -> None:
    # redacts keys in the json object that are in the REDACTED_KEYS set
    # acts on the object in place
    if isinstance(json_obj, dict):
        redacted_keys = {
            k.lower()
            for k in BASE_REDACTED_LOG_KEYS + config.additional_redacted_log_keys + secret_fields
        }
        for key, value in json_obj.items():
            if isinstance(value, dict):
                redact_json_obj(value, secret_fields)
            elif isinstance(value, list):
                for item in value:
                    redact_json_obj(item, secret_fields)
            elif isinstance(value, str):
                if key.lower() in redacted_keys:
                    json_obj[key] = "REDACTED"

    elif isinstance(json_obj, list):
        for item in json_obj:
            redact_json_obj(item, secret_fields)



from enum import Enum

from connector.generated import EntitlementType, ResourceType


class {pascal}ResourceTypes(str, Enum):
    # The global resource is implicit
    pass


class {pascal}EntitlementTypes(str, Enum):
    ROLE = "ROLE"


resource_types: list[ResourceType] = []

entitlement_types: list[EntitlementType] = [
    EntitlementType(
        type_id={pascal}EntitlementTypes.ROLE,
        type_label="Role",
        resource_type_id="",  # means that this entitlement is associated with the global resource
        min=0,
        # You can also set a max, if users can't have infinite of these entitlements
        # max=1,
    )
]



"""
Rate limiting utilities for connector SDK.

This module provides classes and functions to handle rate limiting when making
requests to third-party APIs. It supports batching requests, respecting rate limits,
and handling rate limit headers.
"""

import asyncio
import logging
import time
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any, Generic, TypeVar

import httpx

RequestType = TypeVar("RequestType")  # Input request type
ResponseType = TypeVar("ResponseType")  # Response type

logger = logging.getLogger(__name__)

REQUESTS_PER_WINDOW_CEILING = 0.2
LIMIT_CEILING = 0.2
MAXIMUM_RETRIES = 5

STATIC_RATE_LIMIT_DICTIONARY = [
    "rate limit exceeded",
    "too many requests",
    "quota exceeded",
    "exceeded your rate limit",
    "request limit reached",
]


class RateLimitStrategy(Enum):
    """
    Strategy setting for handling rate limits.

    FIXED - Fixed rate limiting based on predefined limits
    ADAPTIVE - Adaptive rate limiting based on response headers/etc.
    """

    FIXED = "fixed"
    ADAPTIVE = "adaptive"


@dataclass
class RateLimitExtractorResponse:
    """Response from a rate limit extractor."""

    # Remaining requests in the current time window
    remaining: int

    # Total requests allowed in the current time window
    limit: int

    # Reset time in seconds (from the API if available)
    reset: int | None = None

    # Time window in seconds config (from the API if available)
    window_seconds: int | None = None

    # Observed requests (from the API if available)
    observed: str | None = None

    # Requests per window directly config (from the API if available)
    requests_per_window: int | None = None


@dataclass
class RateLimitConfig:
    """Configuration for rate limiting."""

    # App ID
    app_id: str

    # Maximum number of requests per time window
    requests_per_window: int

    # Time window in seconds
    window_seconds: int

    # Maximum retries
    maximum_retries: int = MAXIMUM_RETRIES

    # Strategy for rate limiting
    strategy: RateLimitStrategy = RateLimitStrategy.FIXED

    # Maximum batch size for requests
    max_batch_size: int | None = None

    # Function to extract rate limit info from response
    rate_limit_extractor: Callable[[Any], RateLimitExtractorResponse] | None = None

    # Initial delay between batches in seconds
    initial_delay: float = 0.0

    # Maximum delay between batches in seconds
    max_delay: float = 60.0

    # Backoff factor for exponential backoff
    backoff_factor: float = 1.5

    @classmethod
    def default(cls, app_id: str) -> "RateLimitConfig":
        """Get the default rate limit config."""
        return cls(
            app_id=app_id,
            requests_per_window=30,
            window_seconds=60,
            strategy=RateLimitStrategy.FIXED,
            max_batch_size=15,
        )

    def overwrite(self, **kwargs: Any) -> None:
        """Overwrite the default values with the provided kwargs."""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)


class RateLimitExtractor(ABC):
    """Abstract base class for extracting rate limit information from response."""

    @abstractmethod
    def extract(self, response: Any) -> RateLimitExtractorResponse:
        pass


class RateLimiter(Generic[RequestType, ResponseType]):
    """
    Rate limiter for API requests.

    This class helps manage the rate of requests to third-party APIs by batching
    requests, respecting rate limits, and handling rate limit headers.
    """

    def __init__(self, config: RateLimitConfig):
        self.config = config
        self.request_times: list[float] = []
        self.current_delay = config.initial_delay

        # Set default max_batch_size if not provided
        if self.config.max_batch_size is None:
            self.config.max_batch_size = self.config.requests_per_window

    def _update_request_times(self) -> None:
        """Update the list of request times, removing expired ones."""
        current_time = time.time()
        self.request_times = [
            t for t in self.request_times if current_time - t < self.config.window_seconds
        ]

    def _can_make_request(self) -> bool:
        """Check if a request can be made without exceeding the rate limit."""
        self._update_request_times()
        return len(self.request_times) < self.config.requests_per_window

    def _wait_time_needed(self) -> float:
        """Calculate the time to wait before making the next request."""
        if self._can_make_request():
            return 0.0

        self._update_request_times()
        if not self.request_times:
            return 0.0

        oldest_request = min(self.request_times)
        return oldest_request + self.config.window_seconds - time.time()

    def _update_rate_limits(self, response: Any) -> None:
        """
        Update rate limits based on response.
        """
        if (
            self.config.strategy == RateLimitStrategy.ADAPTIVE
            and self.config.rate_limit_extractor is not None
        ):
            rate_limit_info = self.config.rate_limit_extractor(response)

            # Update the configuration based on the rate limit information
            if rate_limit_info.requests_per_window is not None:
                self.config.requests_per_window = rate_limit_info.requests_per_window

            if rate_limit_info.window_seconds is not None:
                self.config.window_seconds = rate_limit_info.window_seconds

            # If reset time is available, check it
            if rate_limit_info.reset is not None:
                current_time = time.time()
                time_until_reset = rate_limit_info.reset - current_time

                # If reset is within the max delay, adjust the delay
                if (
                    time_until_reset > 0
                    and time_until_reset < self.config.max_delay
                    and rate_limit_info.remaining < rate_limit_info.limit * LIMIT_CEILING
                ):
                    self.current_delay = time_until_reset
                    return

            # If we're approaching the limit, increase the delay
            if (
                rate_limit_info.remaining
                < self.config.requests_per_window * REQUESTS_PER_WINDOW_CEILING
            ):
                self.current_delay = min(
                    self.current_delay * self.config.backoff_factor, self.config.max_delay
                )
            else:
                # If we're not close to the limit, reduce the delay
                self.current_delay = max(
                    self.current_delay / self.config.backoff_factor, self.config.initial_delay
                )

            self.debug_log(
                f"Updated rate limits: {self.config.requests_per_window} requests per {self.config.window_seconds} seconds, current delay: {self.current_delay}"
            )

    def _handle_rate_limit_exceeded(self) -> None:
        """Handle the case when rate limit is exceeded."""
        if self.config.strategy == RateLimitStrategy.FIXED:
            # If on FIXED strategy, we set the current delay to the configured window seconds
            # Since we cannot adapt to the API's rate limits and this is a static way to handle it
            self.debug_log(
                f"Rate limit exceeded; setting delay to {self.config.window_seconds} seconds"
            )
            self.current_delay = self.config.window_seconds

        self.current_delay = min(
            self.current_delay * self.config.backoff_factor, self.config.max_delay
        )

    def debug_log(self, message: str):
        """Log the current rate limit status."""
        print(f"[RateLimiter/{self.config.app_id}] {message}")

    async def execute_requests(
        self,
        requests: list[RequestType],
        request_handler: Callable[[RequestType], ResponseType | asyncio.Future[ResponseType]],
    ) -> list[ResponseType]:
        """
        Execute requests with rate limiting.
        """
        responses: list[ResponseType] = []

        # Handle empty requests list
        if not requests:
            return responses

        # Calc batch size
        batch_size = max(
            1,
            min(
                self.config.max_batch_size or self.config.requests_per_window,
                self.config.requests_per_window,
            ),
        )

        # Process requests in batches
        self.debug_log(f"Executing {len(requests)} requests in {batch_size} batches")

        for i in range(0, len(requests), batch_size):
            batch = requests[i : i + batch_size]
            batch_responses = []

            for request in batch:
                retry_count = 0

                while retry_count <= self.config.maximum_retries:
                    wait_time = self._wait_time_needed()
                    if wait_time > 0:
                        self.debug_log(f"Waiting {wait_time} seconds before making request")
                        await asyncio.sleep(wait_time)

                    try:
                        result = request_handler(request)

                        # Handle both synchronous and asynchronous results
                        if asyncio.iscoroutine(result) or isinstance(result, asyncio.Future):
                            response = await result
                        else:
                            response = result

                        # Record the time
                        self.request_times.append(time.time())

                        # Update rate limits
                        self._update_rate_limits(response)

                        # Add response to batch responses
                        batch_responses.append(response)
                        self.debug_log(
                            f"Request {len(responses) + len(batch_responses)} of {len(requests)} completed"
                        )
                        break  # Success, exit retry loop

                    except Exception as e:
                        # Handle rate limit errors
                        error_str = str(e).lower()
                        if any(
                            keyword in error_str for keyword in STATIC_RATE_LIMIT_DICTIONARY
                        ) or (
                            isinstance(e, httpx.HTTPStatusError)
                            and e.response.status_code == httpx.codes.TOO_MANY_REQUESTS
                        ):
                            self._handle_rate_limit_exceeded()
                            self.debug_log(
                                f"Rate limit exceeded; current delay: {self.current_delay}"
                            )

                            retry_count += 1
                            if retry_count > self.config.maximum_retries:
                                raise Exception(
                                    f"Maximum retries ({self.config.maximum_retries}) reached"
                                ) from e

                            # Retry the request after a delay
                            await asyncio.sleep(self.current_delay)
                        else:
                            # Re-raise other exceptions
                            raise

            responses.extend(batch_responses)

            # Wait between batches
            if i + batch_size < len(requests) and self.current_delay > 0:
                await asyncio.sleep(self.current_delay)

        return responses



from connector.generated import CreatableAccount
from pydantic import Field


# TODO: add/remove fields required to create account
class CreateAccount(CreatableAccount):
    email: str = Field(default=None, description="The email address for the new account")



# PLEASE DO NOT MODIFY THIS FILE MANUALLY.
# Any time someone changes the OpenAPI spec, this file needs to be regenerated. Please
# run: `inv openapi all` to regenerate!


from connector_sdk_types.generated.models.account_status import AccountStatus
from connector_sdk_types.generated.models.account_type import AccountType
from connector_sdk_types.generated.models.activate_account import ActivateAccount
from connector_sdk_types.generated.models.activate_account200_response import (
    ActivateAccount200Response,
)
from connector_sdk_types.generated.models.activate_account_request import ActivateAccountRequest
from connector_sdk_types.generated.models.activate_account_response import ActivateAccountResponse
from connector_sdk_types.generated.models.activated_account import ActivatedAccount
from connector_sdk_types.generated.models.activity_event_type import ActivityEventType
from connector_sdk_types.generated.models.activity_record import ActivityRecord
from connector_sdk_types.generated.models.activity_record_actor import ActivityRecordActor
from connector_sdk_types.generated.models.activity_record_entitlement import (
    ActivityRecordEntitlement,
)
from connector_sdk_types.generated.models.activity_record_target import ActivityRecordTarget
from connector_sdk_types.generated.models.amount import Amount
from connector_sdk_types.generated.models.app_category import AppCategory
from connector_sdk_types.generated.models.app_info import AppInfo
from connector_sdk_types.generated.models.app_info200_response import AppInfo200Response
from connector_sdk_types.generated.models.app_info_request import AppInfoRequest
from connector_sdk_types.generated.models.app_info_request_payload import AppInfoRequestPayload
from connector_sdk_types.generated.models.app_info_response import AppInfoResponse
from connector_sdk_types.generated.models.assign_entitlement import AssignEntitlement
from connector_sdk_types.generated.models.assign_entitlement200_response import (
    AssignEntitlement200Response,
)
from connector_sdk_types.generated.models.assign_entitlement_request import AssignEntitlementRequest
from connector_sdk_types.generated.models.assign_entitlement_response import (
    AssignEntitlementResponse,
)
from connector_sdk_types.generated.models.assigned_entitlement import AssignedEntitlement
from connector_sdk_types.generated.models.auth_credential import AuthCredential
from connector_sdk_types.generated.models.auth_model import AuthModel
from connector_sdk_types.generated.models.authorization_url import AuthorizationUrl
from connector_sdk_types.generated.models.basic_authentication import BasicAuthentication
from connector_sdk_types.generated.models.basic_credential import BasicCredential
from connector_sdk_types.generated.models.capability_schema import CapabilitySchema
from connector_sdk_types.generated.models.creatable_account import CreatableAccount
from connector_sdk_types.generated.models.create_account import CreateAccount
from connector_sdk_types.generated.models.create_account200_response import CreateAccount200Response
from connector_sdk_types.generated.models.create_account_entitlement import CreateAccountEntitlement
from connector_sdk_types.generated.models.create_account_request import CreateAccountRequest
from connector_sdk_types.generated.models.create_account_response import CreateAccountResponse
from connector_sdk_types.generated.models.created_account import CreatedAccount
from connector_sdk_types.generated.models.credential_config import CredentialConfig
from connector_sdk_types.generated.models.custom_attribute_customized_type import (
    CustomAttributeCustomizedType,
)
from connector_sdk_types.generated.models.custom_attribute_schema import CustomAttributeSchema
from connector_sdk_types.generated.models.custom_attribute_type import CustomAttributeType
from connector_sdk_types.generated.models.deactivate_account import DeactivateAccount
from connector_sdk_types.generated.models.deactivate_account200_response import (
    DeactivateAccount200Response,
)
from connector_sdk_types.generated.models.deactivate_account_request import DeactivateAccountRequest
from connector_sdk_types.generated.models.deactivate_account_response import (
    DeactivateAccountResponse,
)
from connector_sdk_types.generated.models.deactivated_account import DeactivatedAccount
from connector_sdk_types.generated.models.delete_account import DeleteAccount
from connector_sdk_types.generated.models.delete_account200_response import DeleteAccount200Response
from connector_sdk_types.generated.models.delete_account_request import DeleteAccountRequest
from connector_sdk_types.generated.models.delete_account_response import DeleteAccountResponse
from connector_sdk_types.generated.models.deleted_account import DeletedAccount
from connector_sdk_types.generated.models.downgrade_license import DowngradeLicense
from connector_sdk_types.generated.models.downgrade_license200_response import (
    DowngradeLicense200Response,
)
from connector_sdk_types.generated.models.downgrade_license_request import DowngradeLicenseRequest
from connector_sdk_types.generated.models.downgrade_license_response import DowngradeLicenseResponse
from connector_sdk_types.generated.models.downgraded_license import DowngradedLicense
from connector_sdk_types.generated.models.entitlement_requirement import EntitlementRequirement
from connector_sdk_types.generated.models.entitlement_type import EntitlementType
from connector_sdk_types.generated.models.error import Error
from connector_sdk_types.generated.models.error_code import ErrorCode
from connector_sdk_types.generated.models.error_response import ErrorResponse
from connector_sdk_types.generated.models.expense import Expense
from connector_sdk_types.generated.models.expense_approval_status import ExpenseApprovalStatus
from connector_sdk_types.generated.models.expense_filters import ExpenseFilters
from connector_sdk_types.generated.models.expense_payment_status import ExpensePaymentStatus
from connector_sdk_types.generated.models.expense_type import ExpenseType
from connector_sdk_types.generated.models.find_entitlement_associations import (
    FindEntitlementAssociations,
)
from connector_sdk_types.generated.models.find_entitlement_associations200_response import (
    FindEntitlementAssociations200Response,
)
from connector_sdk_types.generated.models.find_entitlement_associations_request import (
    FindEntitlementAssociationsRequest,
)
from connector_sdk_types.generated.models.find_entitlement_associations_response import (
    FindEntitlementAssociationsResponse,
)
from connector_sdk_types.generated.models.found_account_data import FoundAccountData
from connector_sdk_types.generated.models.found_entitlement_association import (
    FoundEntitlementAssociation,
)
from connector_sdk_types.generated.models.found_entitlement_data import FoundEntitlementData
from connector_sdk_types.generated.models.found_resource_data import FoundResourceData
from connector_sdk_types.generated.models.get_authorization_url import GetAuthorizationUrl
from connector_sdk_types.generated.models.get_authorization_url200_response import (
    GetAuthorizationUrl200Response,
)
from connector_sdk_types.generated.models.get_authorization_url_request import (
    GetAuthorizationUrlRequest,
)
from connector_sdk_types.generated.models.get_authorization_url_response import (
    GetAuthorizationUrlResponse,
)
from connector_sdk_types.generated.models.get_connected_info import GetConnectedInfo
from connector_sdk_types.generated.models.get_connected_info200_response import (
    GetConnectedInfo200Response,
)
from connector_sdk_types.generated.models.get_connected_info_request import GetConnectedInfoRequest
from connector_sdk_types.generated.models.get_connected_info_response import (
    GetConnectedInfoResponse,
)
from connector_sdk_types.generated.models.get_last_activity import GetLastActivity
from connector_sdk_types.generated.models.get_last_activity200_response import (
    GetLastActivity200Response,
)
from connector_sdk_types.generated.models.get_last_activity_request import GetLastActivityRequest
from connector_sdk_types.generated.models.get_last_activity_response import GetLastActivityResponse
from connector_sdk_types.generated.models.handle_authorization_callback import (
    HandleAuthorizationCallback,
)
from connector_sdk_types.generated.models.handle_authorization_callback200_response import (
    HandleAuthorizationCallback200Response,
)
from connector_sdk_types.generated.models.handle_authorization_callback_request import (
    HandleAuthorizationCallbackRequest,
)
from connector_sdk_types.generated.models.handle_authorization_callback_response import (
    HandleAuthorizationCallbackResponse,
)
from connector_sdk_types.generated.models.handle_client_credentials import HandleClientCredentials
from connector_sdk_types.generated.models.handle_client_credentials_request import (
    HandleClientCredentialsRequest,
)
from connector_sdk_types.generated.models.handle_client_credentials_request200_response import (
    HandleClientCredentialsRequest200Response,
)
from connector_sdk_types.generated.models.handle_client_credentials_response import (
    HandleClientCredentialsResponse,
)
from connector_sdk_types.generated.models.info import Info
from connector_sdk_types.generated.models.info200_response import Info200Response
from connector_sdk_types.generated.models.info_response import InfoResponse
from connector_sdk_types.generated.models.jwt_claims import JWTClaims
from connector_sdk_types.generated.models.jwt_credential import JWTCredential
from connector_sdk_types.generated.models.jwt_headers import JWTHeaders
from connector_sdk_types.generated.models.last_activity_data import LastActivityData
from connector_sdk_types.generated.models.list_accounts import ListAccounts
from connector_sdk_types.generated.models.list_accounts200_response import ListAccounts200Response
from connector_sdk_types.generated.models.list_accounts_request import ListAccountsRequest
from connector_sdk_types.generated.models.list_accounts_response import ListAccountsResponse
from connector_sdk_types.generated.models.list_activity_records import ListActivityRecords
from connector_sdk_types.generated.models.list_activity_records200_response import (
    ListActivityRecords200Response,
)
from connector_sdk_types.generated.models.list_activity_records_request import (
    ListActivityRecordsRequest,
)
from connector_sdk_types.generated.models.list_activity_records_response import (
    ListActivityRecordsResponse,
)
from connector_sdk_types.generated.models.list_connector_app_ids200_response import (
    ListConnectorAppIds200Response,
)
from connector_sdk_types.generated.models.list_custom_attributes_schema import (
    ListCustomAttributesSchema,
)
from connector_sdk_types.generated.models.list_custom_attributes_schema200_response import (
    ListCustomAttributesSchema200Response,
)
from connector_sdk_types.generated.models.list_custom_attributes_schema_request import (
    ListCustomAttributesSchemaRequest,
)
from connector_sdk_types.generated.models.list_custom_attributes_schema_response import (
    ListCustomAttributesSchemaResponse,
)
from connector_sdk_types.generated.models.list_entitlements import ListEntitlements
from connector_sdk_types.generated.models.list_entitlements200_response import (
    ListEntitlements200Response,
)
from connector_sdk_types.generated.models.list_entitlements_request import ListEntitlementsRequest
from connector_sdk_types.generated.models.list_entitlements_response import ListEntitlementsResponse
from connector_sdk_types.generated.models.list_expenses import ListExpenses
from connector_sdk_types.generated.models.list_expenses200_response import ListExpenses200Response
from connector_sdk_types.generated.models.list_expenses_request import ListExpensesRequest
from connector_sdk_types.generated.models.list_expenses_response import ListExpensesResponse
from connector_sdk_types.generated.models.list_resources import ListResources
from connector_sdk_types.generated.models.list_resources200_response import ListResources200Response
from connector_sdk_types.generated.models.list_resources_request import ListResourcesRequest
from connector_sdk_types.generated.models.list_resources_response import ListResourcesResponse
from connector_sdk_types.generated.models.normalized_expense_approval_status import (
    NormalizedExpenseApprovalStatus,
)
from connector_sdk_types.generated.models.normalized_expense_payment_status import (
    NormalizedExpensePaymentStatus,
)
from connector_sdk_types.generated.models.o_auth1_credential import OAuth1Credential
from connector_sdk_types.generated.models.o_auth_authentication import OAuthAuthentication
from connector_sdk_types.generated.models.o_auth_authorization import OAuthAuthorization
from connector_sdk_types.generated.models.o_auth_client_credential import OAuthClientCredential
from connector_sdk_types.generated.models.o_auth_client_credential_authentication import (
    OAuthClientCredentialAuthentication,
)
from connector_sdk_types.generated.models.o_auth_client_credential_authorization import (
    OAuthClientCredentialAuthorization,
)
from connector_sdk_types.generated.models.o_auth_credential import OAuthCredential
from connector_sdk_types.generated.models.o_auth_scopes import OAuthScopes
from connector_sdk_types.generated.models.oauth_credentials import OauthCredentials
from connector_sdk_types.generated.models.page import Page
from connector_sdk_types.generated.models.refresh_access_token import RefreshAccessToken
from connector_sdk_types.generated.models.refresh_access_token200_response import (
    RefreshAccessToken200Response,
)
from connector_sdk_types.generated.models.refresh_access_token_request import (
    RefreshAccessTokenRequest,
)
from connector_sdk_types.generated.models.refresh_access_token_response import (
    RefreshAccessTokenResponse,
)
from connector_sdk_types.generated.models.release_resources import ReleaseResources
from connector_sdk_types.generated.models.release_resources200_response import (
    ReleaseResources200Response,
)
from connector_sdk_types.generated.models.release_resources_request import ReleaseResourcesRequest
from connector_sdk_types.generated.models.release_resources_response import ReleaseResourcesResponse
from connector_sdk_types.generated.models.release_resources_status import ReleaseResourcesStatus
from connector_sdk_types.generated.models.resource_type import ResourceType
from connector_sdk_types.generated.models.service_account_credential import ServiceAccountCredential
from connector_sdk_types.generated.models.service_account_type import ServiceAccountType
from connector_sdk_types.generated.models.spend_user import SpendUser
from connector_sdk_types.generated.models.standard_capability_name import StandardCapabilityName
from connector_sdk_types.generated.models.time_range import TimeRange
from connector_sdk_types.generated.models.token_authentication import TokenAuthentication
from connector_sdk_types.generated.models.token_credential import TokenCredential
from connector_sdk_types.generated.models.token_type import TokenType
from connector_sdk_types.generated.models.transfer_data import TransferData
from connector_sdk_types.generated.models.transfer_data200_response import TransferData200Response
from connector_sdk_types.generated.models.transfer_data_request import TransferDataRequest
from connector_sdk_types.generated.models.transfer_data_response import TransferDataResponse
from connector_sdk_types.generated.models.transfer_data_status import TransferDataStatus
from connector_sdk_types.generated.models.unassign_entitlement import UnassignEntitlement
from connector_sdk_types.generated.models.unassign_entitlement200_response import (
    UnassignEntitlement200Response,
)
from connector_sdk_types.generated.models.unassign_entitlement_request import (
    UnassignEntitlementRequest,
)
from connector_sdk_types.generated.models.unassign_entitlement_response import (
    UnassignEntitlementResponse,
)
from connector_sdk_types.generated.models.unassigned_entitlement import UnassignedEntitlement
from connector_sdk_types.generated.models.update_account200_response import UpdateAccount200Response
from connector_sdk_types.generated.models.update_account_request import UpdateAccountRequest
from connector_sdk_types.generated.models.update_account_response import UpdateAccountResponse
from connector_sdk_types.generated.models.updateable_account import UpdateableAccount
from connector_sdk_types.generated.models.validate_credentials import ValidateCredentials
from connector_sdk_types.generated.models.validate_credentials200_response import (
    ValidateCredentials200Response,
)
from connector_sdk_types.generated.models.validate_credentials_request import (
    ValidateCredentialsRequest,
)
from connector_sdk_types.generated.models.validate_credentials_response import (
    ValidateCredentialsResponse,
)
from connector_sdk_types.generated.models.validated_credentials import ValidatedCredentials
from connector_sdk_types.generated.models.vendor import Vendor

__all__ = [
    "AccountStatus",
    "AccountType",
    "ActivateAccount",
    "ActivateAccount200Response",
    "ActivateAccountRequest",
    "ActivateAccountResponse",
    "ActivatedAccount",
    "ActivityEventType",
    "ActivityRecord",
    "ActivityRecordActor",
    "ActivityRecordEntitlement",
    "ActivityRecordTarget",
    "Amount",
    "AppCategory",
    "AppInfo",
    "AppInfo200Response",
    "AppInfoRequest",
    "AppInfoRequestPayload",
    "AppInfoResponse",
    "AssignEntitlement",
    "AssignEntitlement200Response",
    "AssignEntitlementRequest",
    "AssignEntitlementResponse",
    "AssignedEntitlement",
    "AuthCredential",
    "AuthModel",
    "AuthorizationUrl",
    "BasicAuthentication",
    "BasicCredential",
    "CapabilitySchema",
    "CreatableAccount",
    "CreateAccount",
    "CreateAccount200Response",
    "CreateAccountEntitlement",
    "CreateAccountRequest",
    "CreateAccountResponse",
    "CreatedAccount",
    "CredentialConfig",
    "CustomAttributeCustomizedType",
    "CustomAttributeSchema",
    "CustomAttributeType",
    "DeactivateAccount",
    "DeactivateAccount200Response",
    "DeactivateAccountRequest",
    "DeactivateAccountResponse",
    "DeactivatedAccount",
    "DeleteAccount",
    "DeleteAccount200Response",
    "DeleteAccountRequest",
    "DeleteAccountResponse",
    "DeletedAccount",
    "DowngradeLicense",
    "DowngradeLicense200Response",
    "DowngradeLicenseRequest",
    "DowngradeLicenseResponse",
    "DowngradedLicense",
    "EntitlementRequirement",
    "EntitlementType",
    "Error",
    "ErrorCode",
    "ErrorResponse",
    "Expense",
    "ExpenseApprovalStatus",
    "ExpenseFilters",
    "ExpensePaymentStatus",
    "ExpenseType",
    "FindEntitlementAssociations",
    "FindEntitlementAssociations200Response",
    "FindEntitlementAssociationsRequest",
    "FindEntitlementAssociationsResponse",
    "FoundAccountData",
    "FoundEntitlementAssociation",
    "FoundEntitlementData",
    "FoundResourceData",
    "GetAuthorizationUrl",
    "GetAuthorizationUrl200Response",
    "GetAuthorizationUrlRequest",
    "GetAuthorizationUrlResponse",
    "GetConnectedInfo",
    "GetConnectedInfo200Response",
    "GetConnectedInfoRequest",
    "GetConnectedInfoResponse",
    "GetLastActivity",
    "GetLastActivity200Response",
    "GetLastActivityRequest",
    "GetLastActivityResponse",
    "HandleAuthorizationCallback",
    "HandleAuthorizationCallback200Response",
    "HandleAuthorizationCallbackRequest",
    "HandleAuthorizationCallbackResponse",
    "HandleClientCredentials",
    "HandleClientCredentialsRequest",
    "HandleClientCredentialsRequest200Response",
    "HandleClientCredentialsResponse",
    "Info",
    "Info200Response",
    "InfoResponse",
    "JWTClaims",
    "JWTCredential",
    "JWTHeaders",
    "LastActivityData",
    "ListAccounts",
    "ListAccounts200Response",
    "ListAccountsRequest",
    "ListAccountsResponse",
    "ListActivityRecords",
    "ListActivityRecords200Response",
    "ListActivityRecordsRequest",
    "ListActivityRecordsResponse",
    "ListConnectorAppIds200Response",
    "ListCustomAttributesSchema",
    "ListCustomAttributesSchema200Response",
    "ListCustomAttributesSchemaRequest",
    "ListCustomAttributesSchemaResponse",
    "ListEntitlements",
    "ListEntitlements200Response",
    "ListEntitlementsRequest",
    "ListEntitlementsResponse",
    "ListExpenses",
    "ListExpenses200Response",
    "ListExpensesRequest",
    "ListExpensesResponse",
    "ListResources",
    "ListResources200Response",
    "ListResourcesRequest",
    "ListResourcesResponse",
    "NormalizedExpenseApprovalStatus",
    "NormalizedExpensePaymentStatus",
    "OAuth1Credential",
    "OAuthAuthentication",
    "OAuthAuthorization",
    "OAuthClientCredential",
    "OAuthClientCredentialAuthentication",
    "OAuthClientCredentialAuthorization",
    "OAuthCredential",
    "OAuthScopes",
    "OauthCredentials",
    "Page",
    "RefreshAccessToken",
    "RefreshAccessToken200Response",
    "RefreshAccessTokenRequest",
    "RefreshAccessTokenResponse",
    "ReleaseResources",
    "ReleaseResources200Response",
    "ReleaseResourcesRequest",
    "ReleaseResourcesResponse",
    "ReleaseResourcesStatus",
    "ResourceType",
    "ServiceAccountCredential",
    "ServiceAccountType",
    "SpendUser",
    "StandardCapabilityName",
    "TimeRange",
    "TokenAuthentication",
    "TokenCredential",
    "TokenType",
    "TransferData",
    "TransferData200Response",
    "TransferDataRequest",
    "TransferDataResponse",
    "TransferDataStatus",
    "UnassignEntitlement",
    "UnassignEntitlement200Response",
    "UnassignEntitlementRequest",
    "UnassignEntitlementResponse",
    "UnassignedEntitlement",
    "UpdateAccount200Response",
    "UpdateAccountRequest",
    "UpdateAccountResponse",
    "UpdateableAccount",
    "ValidateCredentials",
    "ValidateCredentials200Response",
    "ValidateCredentialsRequest",
    "ValidateCredentialsResponse",
    "ValidatedCredentials",
    "Vendor",
]



from __future__ import annotations

from pydantic import BaseModel, ValidationError
from pydantic.fields import FieldInfo


def _get_model_field_info(model: type[BaseModel], field_name: str) -> FieldInfo | None:
    return model.model_fields.get(field_name) if hasattr(model, "model_fields") else None


def _extract_field_title(model: type[BaseModel], field_name: str) -> str:
    field = _get_model_field_info(model, field_name)
    if field is None:
        return field_name

    return field.title or field_name


def get_missing_field_titles(
    error: ValidationError, model: type[BaseModel]
) -> tuple[list[str], bool]:
    """Return a list of titles for fields reported as missing."""

    titles: list[str] = []
    other_errors = False

    for err in error.errors():
        if err.get("type") != "missing":
            other_errors = True
            continue

        loc = err.get("loc", ())
        if not loc:
            continue

        # the last part of the location tuple is the field name
        field_name: str | None = None
        for part in reversed(loc):
            if isinstance(part, str):
                field_name = part
                break
        if field_name is None:
            continue

        title = _extract_field_title(model, field_name)
        titles.append(title)

    return titles, other_errors



import logging
import typing as t
from abc import abstractmethod
from collections.abc import Callable
from pathlib import Path
from typing import Any

from connector_sdk_types.generated import ErrorCode
from gql import Client
from gql.client import AsyncClientSession
from gql.dsl import DSLSchema
from graphql import GraphQLSchema, build_schema
from httpx import Response
from typing_extensions import Self

from connector.httpx_rewrite import AsyncClient
from connector.oai.capability import Request
from connector.oai.errors import ConnectorError
from connector.utils.rate_limiting import RateLimitConfig, RateLimiter

logger = logging.getLogger(__name__)

# Type alias for batch requests
BatchRequest = tuple[tuple[Any, ...], dict[str, Any]]
BatchRequests = list[BatchRequest]


class RateLimitedClient(AsyncClient):
    """A wrapper around httpx.AsyncClient that applies rate limiting."""

    def __init__(self, base_client: AsyncClient, rate_limit_config: RateLimitConfig):
        self.base_client = base_client
        self.rate_limiter = RateLimiter[Callable[[], Any], Any](rate_limit_config)

    def __getattr__(self, name):
        """Delegate attribute access to the underlying base_client."""
        return getattr(self.base_client, name)

    async def _execute_request(self, method, *args, **kwargs) -> Response:
        """Execute a request with rate limiting."""

        async def request_func():
            func = getattr(self.base_client, method)
            return await func(*args, **kwargs)

        # We're only sending one request, but using the batch API
        responses = await self.rate_limiter.execute_requests([request_func], lambda x: x())
        if responses:
            return responses[0]

        raise ConnectorError(
            message="No response from the API",
            error_code=ErrorCode.API_ERROR,
        )

    async def batch_request(
        self, method: str, requests: BatchRequests, **default_kwargs
    ) -> list[Response]:
        """
        Execute multiple requests of the same HTTP method in a batch with rate limiting.

        Args:
            method: The HTTP method to use (e.g., 'get', 'post', 'patch', 'put', 'delete')
            requests: List of tuples containing (args, kwargs) for each request
            **default_kwargs: Default keyword arguments to apply to all requests

        Returns:
            List of Response objects in the same order as the input requests

        Example:
            # Prepare batch requests
            batch_requests: BatchRequests = [
                ((f"{Endpoint.USERS}/{user_id}",), {}) for user_id in user_ids
            ]

            # Execute batch request
            responses = await self.batch_request("get", batch_requests)

            # Convert responses to User objects
            users = []
            for response in responses:
                user = create_client_response(response, User)
                users.append(user)

            return users
        """
        if not requests:
            return []

        # Create request functions for each request
        request_functions = []
        for args, kwargs in requests:
            # Create a closure to capture the current args and kwargs
            def make_request_func(request_args=args, request_kwargs=kwargs):
                async def execute_single_request():
                    # Merge default kwargs with request-specific kwargs
                    merged_kwargs = {**default_kwargs, **request_kwargs}
                    func = getattr(self.base_client, method)
                    return await func(*request_args, **merged_kwargs)

                return execute_single_request

            request_functions.append(make_request_func())

        # Execute all requests with rate limiting
        responses = await self.rate_limiter.execute_requests(request_functions, lambda x: x())

        return responses

    def get_state(self) -> tuple[RateLimitConfig, float]:
        """
        Get the current rate limit state.

        Returns a tuple of the rate limit config and the current delay.
        """
        return self.rate_limiter.config, self.rate_limiter.current_delay

    async def get(self, *args, **kwargs):
        return await self._execute_request("get", *args, **kwargs)

    async def post(self, *args, **kwargs):
        return await self._execute_request("post", *args, **kwargs)

    async def patch(self, *args, **kwargs):
        return await self._execute_request("patch", *args, **kwargs)

    async def put(self, *args, **kwargs):
        return await self._execute_request("put", *args, **kwargs)

    async def delete(self, *args, **kwargs):
        return await self._execute_request("delete", *args, **kwargs)

    # Add async context manager methods
    async def __aenter__(self):
        if hasattr(self.base_client, "__aenter__"):
            await self.base_client.__aenter__()
        return self

    async def __aexit__(self, exc_type=None, exc_val=None, exc_tb=None):
        if hasattr(self.base_client, "__aexit__"):
            await self.base_client.__aexit__(exc_type, exc_val, exc_tb)


class BaseIntegrationClient:
    _http_client: AsyncClient | RateLimitedClient
    _rate_limit_config: RateLimitConfig | None = None

    @classmethod
    @abstractmethod
    def prepare_client_args(cls, args: Request) -> dict[str, t.Any]:
        pass

    @classmethod
    def build_client(cls, args: Request) -> AsyncClient:
        return AsyncClient(**cls.prepare_client_args(args))

    def get_current_rate_limits(self) -> tuple[RateLimitConfig | None, float]:
        """
        Get the current rate limit state.

        Returns a tuple of the rate limit config and the current delay. (or None if the client is not rate limited)
        """
        if isinstance(self._http_client, RateLimitedClient):
            return self._http_client.get_state()
        return None, 0

    async def batch_request(
        self, method: str, requests: BatchRequests, **default_kwargs
    ) -> list[Response]:
        """
        Execute multiple requests of the same HTTP method in a batch with/without rate limiting.

        Args:
            method: The HTTP method to use (e.g., 'get', 'post', 'patch', 'put', 'delete')
            requests: List of tuples containing (args, kwargs) for each request
            **default_kwargs: Default keyword arguments to apply to all requests

        Returns:
            List of Response objects in the same order as the input requests
        """
        if isinstance(self._http_client, RateLimitedClient):
            return await self._http_client.batch_request(method, requests, **default_kwargs)
        else:
            # Fallback for non-rate-limited clients
            responses = []
            for args, kwargs in requests:
                merged_kwargs = {**default_kwargs, **kwargs}
                func = getattr(self._http_client, method)
                response = await func(*args, **merged_kwargs)
                responses.append(response)
            return responses

    def __init__(self, args: Request, rate_limit_config: RateLimitConfig | None = None) -> None:
        base_client = self.build_client(args)
        self._http_client = base_client

        rate_limiting = rate_limit_config or self._rate_limit_config
        if rate_limiting is not None:
            self._http_client = RateLimitedClient(base_client, rate_limiting)

    async def __aenter__(self):
        await self._http_client.__aenter__()
        return self

    async def __aexit__(self, exc_type=None, exc_val=None, exc_tb=None):
        await self._http_client.__aexit__(exc_type, exc_val, exc_tb)
        if exc_val is not None:
            raise exc_val


class BaseGraphQLSession(AsyncClientSession):
    def __init__(self, args: Request):
        super().__init__(client=self.build_client(args))

    async def __aenter__(self) -> Self:
        await self.client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.client.__aexit__(exc_type=exc_type, exc=exc, tb=tb)

        if exc_type is not None:
            raise exc

    @classmethod
    @abstractmethod
    def prepare_client_args(cls, args: Request) -> dict[str, t.Any]:
        pass

    @classmethod
    def build_client(cls, args: Request) -> Client:
        return Client(**cls.prepare_client_args(args))

    @classmethod
    def load_schema(cls, schema_file_path: str | Path) -> GraphQLSchema:
        """Load the GraphQL schema from a file."""
        with open(schema_file_path) as f:
            return build_schema(f.read())

    @property
    def schema(self) -> DSLSchema:
        if self.client.schema is None:
            raise ConnectorError(
                message="Failed to fetch schema",
                error_code=ErrorCode.API_ERROR,
            )

        return DSLSchema(self.client.schema)



import logging
from unittest.mock import MagicMock, patch

import pytest
from connector.handlers.lumos_log_handler import LumosLogHandler


@pytest.fixture
def mock_config():
    with patch("connector.handlers.lumos_log_handler.config") as mock:
        mock.agent_identifier = "test-agent"
        mock.on_prem_agent_api_key = "test-key"
        mock.base_url = "https://test-base-url"
        yield mock


@pytest.fixture
def lumos_handler(mock_config):
    return LumosLogHandler()


def test_lumos_handler_initialization(mock_config):
    handler = LumosLogHandler()
    assert handler.agent_identifier == "test-agent"


def test_lumos_handler_ignores_non_error_logs(lumos_handler):
    record = logging.LogRecord(
        name="test",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="Test message",
        args=(),
        exc_info=None,
    )
    lumos_handler.emit(record)
    # No exception should be raised, and the log should be ignored


def test_lumos_handler_sends_error_logs(lumos_handler, mock_config):
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.raise_for_status = MagicMock()
    mock_client.post.return_value = mock_response

    with patch("connector.handlers.lumos_log_handler.AsyncClient") as mock_async_client:
        mock_async_client.return_value.__aenter__.return_value = mock_client

        record = logging.LogRecord(
            name="test",
            level=logging.ERROR,
            pathname="test.py",
            lineno=1,
            msg="Test error message",
            args=(),
            exc_info=None,
        )

        # Set a formatter to ensure the log message is properly formatted
        lumos_handler.setFormatter(logging.Formatter("%(message)s"))

        # Emit the log record
        lumos_handler.emit(record)

        # Verify the HTTP client was called with correct data
        mock_client.post.assert_called_once()
        call_args = mock_client.post.call_args
        assert call_args[0][0] == "/logs"
        assert call_args[1]["json"] == {
            "agent_identifier": "test-agent",
            "log": "Test error message",
        }


def test_lumos_handler_handles_http_errors(lumos_handler, mock_config):
    mock_client = MagicMock()
    mock_client.post.side_effect = Exception("HTTP Error")

    with patch("connector.handlers.lumos_log_handler.AsyncClient") as mock_async_client:
        mock_async_client.return_value.__aenter__.return_value = mock_client

        record = logging.LogRecord(
            name="test",
            level=logging.ERROR,
            pathname="test.py",
            lineno=1,
            msg="Test error message",
            args=(),
            exc_info=None,
        )

        # The handler should not raise an exception
        lumos_handler.emit(record)

        # Verify the HTTP client was called
        mock_client.post.assert_called_once()


def test_lumos_handler_formatting(mock_config):
    handler = LumosLogHandler()
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)

    record = logging.LogRecord(
        name="test",
        level=logging.ERROR,
        pathname="test.py",
        lineno=1,
        msg="Test message",
        args=(),
        exc_info=None,
    )

    formatted = handler.format(record)
    assert formatted == "Test message"



from connector_sdk_types.generated import Error, ErrorCode, ErrorResponse

from .oai.errors import ConnectorError, DefaultHandler, ExceptionHandler, HTTPHandler

__all__ = [
    "Error",
    "ErrorCode",
    "ErrorResponse",
    "ConnectorError",
    "DefaultHandler",
    "ExceptionHandler",
    "HTTPHandler",
]



# PLEASE DO NOT MODIFY THIS FILE MANUALLY.
# This file re-exports everything from the connector-sdk-types package for backwards compatibility.
# Any time someone changes the OpenAPI spec, this file needs to be regenerated.

from connector_sdk_types.generated.models.list_connector_app_ids200_response import ListConnectorAppIds200Response

__all__ = [
    "ListConnectorAppIds200Response",
]



from enum import Enum
from typing import TYPE_CHECKING, Any, cast

from connector_sdk_types.generated import (
    AppInfo,
    AppInfoRequest,
    AppInfoResponse,
    AuthModel,
    BasicAuthentication,
    CapabilitySchema,
    JWTCredential,
    OAuth1Credential,
    OAuthAuthentication,
    OAuthAuthorization,
    OAuthClientCredentialAuthentication,
    OAuthClientCredentialAuthorization,
    ServiceAccountCredential,
    StandardCapabilityName,
    TokenAuthentication,
)
from pydantic import BaseModel
from typing_extensions import assert_never

from connector.oai.capability import generate_capability_schema
from connector.oai.modules.base_module import BaseIntegrationModule
from connector.oai.modules.oauth_module_types import OAuthConfig, OAuthFlowType, OAuthSettings

if TYPE_CHECKING:
    from connector.oai.integration import CapabilityMetadata, Integration

"""
Some mappings for the info module.
These are used to safely map different auth models to correct types and divide the info output.
"""

AUTH_TYPE_MAPPING = {
    AuthModel.OAUTH: (OAuthAuthentication, OAuthAuthorization),
    AuthModel.OAUTH_CLIENT_CREDENTIALS: (
        OAuthClientCredentialAuthentication,
        OAuthClientCredentialAuthorization,
    ),
    AuthModel.OAUTH1: (OAuth1Credential, None),
    AuthModel.BASIC: (BasicAuthentication, None),
    AuthModel.TOKEN: (TokenAuthentication, None),
    AuthModel.JWT: (JWTCredential, None),
    AuthModel.SERVICE_ACCOUNT: (ServiceAccountCredential, None),
}

OAS_FLOW_TYPE_MAPPING = {
    OAuthFlowType.CLIENT_CREDENTIALS: "clientCredentials",
    OAuthFlowType.CODE_FLOW: "authorizationCode",
}


class InfoModule(BaseIntegrationModule):
    """
    Info module is responsible for generating the OpenAPI specification for the app.

    This module:
    - Registers the `app_info` capability on the app instance
    - Generates the OpenAPI specification for the app
    - Attaches custom information using the OAS extensions
    """

    def __init__(self):
        super().__init__()

    def register(self, integration: "Integration"):
        self.integration = integration

        self.register_info_capability()
        self.add_capability(StandardCapabilityName.APP_INFO.value)

    def register_info_capability(self):
        @self.integration.register_capability(StandardCapabilityName.APP_INFO)
        async def info_capability(args: AppInfoRequest) -> AppInfoResponse:
            """
            This is the `app_info` capability, it returns information about the app.
            This capability accepts optional settings/auth objects, just like if you were to run another capability.
            These optional objects can mutate the connector's "state" and could modify the info response,
            making this the `connected info` capability.
            """
            openapi_spec = self.generate_openapi_spec(args)
            response = AppInfoResponse(
                response=AppInfo(
                    app_id=self.integration.app_id,
                    app_schema=openapi_spec,
                )
            )

            return response

        return info_capability

    """
    Info methods
    """

    def generate_openapi_spec(self, args: AppInfoRequest) -> dict[str, Any]:
        capabilities = self.capability_schema_info()
        authentication, authorization, security_schemas = self.get_auth_schema(args)

        # Create the base specification
        spec: dict[str, Any] = self._create_base_spec()

        # Add security schemes
        spec["components"]["securitySchemes"] = security_schemas

        # Track used components
        used_refs: set[str] = set()

        # Settings schema
        if hasattr(self.integration, "settings_model"):
            self._add_settings_to_schema(
                spec,
                self.integration.settings_model,
                used_refs,
            )

        # Credentials schema
        if not self.integration.auth:
            credentials_schema = {
                "type": "array",
                "items": {
                    "allOf": [
                        {
                            "type": "object",
                            "properties": {
                                "id": {"type": "string", "enum": [cred_id]},
                                auth_type: auth_info,
                            },
                            "required": ["id", auth_type],
                        }
                        for cred_id, auth_info in authentication.items()
                        if "type" in auth_info
                        for auth_type in [self._get_auth_type(auth_info)]
                    ]
                },
            }
        else:
            auth_schema = authentication[self.integration.app_id]

        if not self.integration.auth:
            spec["components"]["schemas"]["Credentials"] = credentials_schema

            # Add the credentials schema to the used refs and process everything
            used_refs.add("Credentials")
            self._collect_refs(credentials_schema, used_refs)
            self._process_definitions(credentials_schema, spec)
        else:
            spec["components"]["schemas"]["Auth"] = auth_schema
            used_refs.add("Auth")
            self._collect_refs(auth_schema, used_refs)
            self._process_definitions(auth_schema, spec)

        # Add paths/operations from capabilities
        for cap_name, cap_schema in capabilities.items():
            # API path
            path = f"/{cap_name.lower()}"

            # Convert Pydantic schemas to dictionaries
            request_schema = self._convert_null_type(cap_schema.argument)
            response_schema = self._convert_null_type(cap_schema.output)

            # Collect references
            self._collect_refs(request_schema, used_refs)
            self._collect_refs(response_schema, used_refs)

            # Setup proper definitions in the specification
            for schema in [request_schema, response_schema]:
                if isinstance(schema, dict):
                    self._process_definitions(schema, spec)

            # Get the processed payload references
            request_payload_schema = request_schema.get("properties", {}).get("request", {})
            response_payload_schema = response_schema.get("properties", {}).get("response", {})

            # Resolve the references
            if "$ref" in request_payload_schema:
                ref_path = request_payload_schema["$ref"].split("/")[-1]
                request_payload_schema = spec["components"]["schemas"][ref_path]
            if "$ref" in response_payload_schema:
                ref_path = response_payload_schema["$ref"].split("/")[-1]
                response_payload_schema = spec["components"]["schemas"][ref_path]

            # Determine operation category
            operation_category = "Capabilities"
            operation_subcategory: str | None = None
            if isinstance(request_payload_schema, dict):
                if "x-capability-category" in request_payload_schema:
                    operation_category = request_payload_schema[
                        "x-capability-category"
                    ].capitalize()
                if "x-capability-level" in request_payload_schema:
                    operation_subcategory = (
                        f"{request_payload_schema['x-capability-level'].capitalize()} Capabilities"
                    )

            # Request body schema
            request_body_schema: dict[str, Any] = {
                "type": "object",
                "properties": {
                    "request": request_payload_schema,
                    "settings": {
                        "$ref": "#/components/schemas/Settings"
                        if hasattr(self.integration, "settings_model")
                        else {"type": "object", "example": "{}"},
                    },
                },
                "required": ["request", "credentials", "settings"],
            }

            # Add credentials schema to request if not an authorization capability
            if operation_category not in ["Authorization"] and not self.integration.auth:
                request_body_schema["properties"]["credentials"] = {
                    "$ref": "#/components/schemas/Credentials"
                }
            elif operation_category not in ["Authorization"] and self.integration.auth:
                request_body_schema["properties"]["auth"] = {"$ref": "#/components/schemas/Auth"}

            # Operation security schemes
            security_requirements = []
            for cred_id in authentication.keys():
                security_scheme_name = cred_id

                if authorization.get(cred_id) and "oauth_settings" in authorization[cred_id]:
                    scopes = authorization[cred_id]["oauth_settings"]["scopes"]
                    if scopes:
                        capability_scopes = list(scopes.keys())
                        security_requirements.append({security_scheme_name: capability_scopes})
                    else:
                        security_requirements.append({security_scheme_name: []})
                else:
                    security_requirements.append({security_scheme_name: []})

            # Tags
            tags = []
            if operation_subcategory:
                tags.append(operation_subcategory)
            else:
                tags.append(operation_category)  # fallback

            # Type annotations for operation
            operation: dict[str, Any] = {
                "operationId": cap_name,
                "tags": tags,
                "summary": cap_schema.display_name,
                "description": cap_schema.description,
                "requestBody": {"content": {"application/json": {"schema": request_body_schema}}},
                "responses": {
                    "200": {
                        "description": "Successful response",
                        "content": {"application/json": {"schema": response_payload_schema}},
                    }
                },
                "security": security_requirements,
            }

            spec["paths"][path] = {"post": operation}

        # Remove unused components
        spec["components"]["schemas"] = {
            name: schema
            for name, schema in spec["components"]["schemas"].items()
            if name in used_refs
        }

        return self._serialize_for_json(spec)

    def capabilities_info(self) -> list[str]:
        return [capability for capability in self.integration.capabilities]

    def capability_schema_info(self) -> dict[str, CapabilitySchema]:
        """Modified paste from Integration"""
        capability_names = sorted(self.integration.capabilities.keys())
        capability_schema: dict[str, CapabilitySchema] = {}
        for capability_name in capability_names:
            command_types = generate_capability_schema(
                impl=self.integration.capabilities[capability_name],
                full_schema=True,
            )

            # Capability metadata
            capability_metadata: CapabilityMetadata | None = None
            if capability_name in self.integration.capability_metadata:
                capability_metadata = self.integration.capability_metadata[capability_name]

            # Display name
            display_name: str | None = capability_name.replace("_", " ").title()
            if (
                capability_metadata is not None
                and capability_metadata.display_name is not None
                and isinstance(capability_metadata.display_name, str)
            ):
                display_name = capability_metadata.display_name

            # Description
            description = None
            if (
                capability_metadata is not None
                and capability_metadata.description is not None
                and isinstance(capability_metadata.description, str)
            ):
                description = capability_metadata.description

            # Without description, get the docstring from the capability method
            if description is None:
                description = self.integration.capabilities[capability_name].__doc__

            capability_schema[capability_name] = CapabilitySchema(
                argument=command_types.argument,
                output=command_types.output,
                display_name=display_name,
                description=description,
            )

        return capability_schema

    def get_auth_schema(
        self, args: AppInfoRequest
    ) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
        # Leaving these here for now, we can use these to mutate the spec
        # settings = get_settings(args, self.integration.settings_model)
        # auth = args.credentials

        authentication = {}
        authorization = {}
        security_schemas = {}

        if self.integration.credentials:
            """
            Forwards support for credentials
            """
            for credential in self.integration.credentials:
                security_schema_name = f"{credential.id}"

                # Model mappings
                (
                    authentication_model,
                    authorization_model,
                ) = AUTH_TYPE_MAPPING[credential.type]

                # Base models
                if credential.input_model:
                    # Custom input model provided by the app
                    authentication[credential.id] = self._get_model_extended_json_schema(
                        cast(type[BaseModel], credential.input_model)
                    )
                else:
                    # Standard input models mapped from the mapping
                    authentication[credential.id] = self._get_model_extended_json_schema(
                        cast(type[BaseModel], authentication_model)
                    )

                if authorization_model:
                    # Not every app has an authorization model
                    authorization[credential.id] = self._get_model_extended_json_schema(
                        cast(type[BaseModel], authorization_model)
                    )

                # Markdown data
                authentication[credential.id]["description"] = credential.description

                # Specific connector OAuth settings
                if isinstance(credential, OAuthConfig) and credential.oauth_settings:
                    oauth = credential.oauth_settings
                    scopes = self.get_oauth_scopes(credential.id, args)

                    token_url = ""
                    authorization_url = ""

                    # Get token URL and authorization URL
                    if callable(oauth.token_url):
                        try:
                            token_url = oauth.token_url(args)
                        except Exception as _:
                            token_url = ""
                    else:
                        token_url = oauth.token_url or ""

                    if callable(oauth.authorization_url):
                        try:
                            authorization_url = oauth.authorization_url(args)
                        except Exception as _:
                            authorization_url = ""
                    else:
                        authorization_url = oauth.authorization_url or ""

                    # Custom object
                    authorization[credential.id]["oauth_settings"] = {
                        "oauth_type": oauth.flow_type,
                        "token_url": token_url,
                        "authorization_url": authorization_url,
                        "scopes": scopes,
                        "pkce_enabled": oauth.pkce,
                    }

                    # Process OAS flows
                    flows: dict[str, Any] = {}
                    flow_type = OAS_FLOW_TYPE_MAPPING[oauth.flow_type]
                    if flow_type == "authorizationCode":
                        flows[flow_type] = {}
                        flows[flow_type]["authorizationUrl"] = authorization_url or ""
                        flows[flow_type]["tokenUrl"] = token_url or ""
                        flows[flow_type]["scopes"] = scopes or {}
                    elif flow_type == "clientCredentials":
                        flows[flow_type] = {}
                        flows[flow_type]["tokenUrl"] = token_url or ""
                        flows[flow_type]["scopes"] = scopes or {}

                    security_schemas[security_schema_name] = {
                        "type": "oauth2",
                        "flows": flows,
                    }
                else:
                    if credential.type == AuthModel.TOKEN or credential.type == AuthModel.JWT:
                        security_schemas[security_schema_name] = {
                            "type": "apiKey",
                            "in": "header",
                            "name": "Authorization",
                            "description": "API Token / API Key for authentication",
                        }
                    elif credential.type == AuthModel.BASIC:
                        security_schemas[security_schema_name] = {
                            "type": "http",
                        }
        elif self.integration.auth:
            """
            Backwards support for auth settings

            Note: This can be removed once 'auth' is phased out and only 'credentials' is used.
            """
            security_schema_name = f"{self.integration.app_id}"
            auth_model = self._get_model_extended_json_schema(self.integration.auth)
            auth_type = AuthModel(auth_model["x-credential-type"])

            authentication[self.integration.app_id] = {
                "properties": {
                    auth_type: {
                        "type": "object",
                        **auth_model,
                    },
                },
                "required": [auth_type],
            }
            integration_oauth_settings: OAuthSettings | None = self.integration.oauth_settings

            match auth_type:
                case AuthModel.OAUTH:
                    authorization_url = ""
                    if integration_oauth_settings is not None:
                        if callable(integration_oauth_settings.authorization_url):
                            try:
                                authorization_url = integration_oauth_settings.authorization_url(
                                    args
                                )
                            except Exception as _:
                                authorization_url = ""
                        else:
                            authorization_url = integration_oauth_settings.authorization_url or ""

                    security_schemas[security_schema_name] = {
                        "type": "oauth2",
                        "flows": {
                            "authorizationCode": {
                                "authorizationUrl": authorization_url,
                            },
                        },
                    }
                case AuthModel.OAUTH_CLIENT_CREDENTIALS:
                    token_url = ""
                    if integration_oauth_settings is not None:
                        if callable(integration_oauth_settings.token_url):
                            try:
                                token_url = integration_oauth_settings.token_url(args)
                            except Exception as _:
                                token_url = ""
                        else:
                            token_url = integration_oauth_settings.token_url or ""

                    security_schemas[security_schema_name] = {
                        "type": "oauth2",
                        "flows": {
                            "clientCredentials": {
                                "tokenUrl": token_url,
                            },
                        },
                    }
                case AuthModel.OAUTH1:
                    security_schemas[security_schema_name] = {
                        "type": "apiKey",
                        "in": "header",
                        "name": "Authorization",
                        "description": "OAuth 1.0a signed request token for authentication",
                    }  # OAS does not support legacy OAuth officially
                case AuthModel.TOKEN:
                    security_schemas[security_schema_name] = {
                        "type": "apiKey",
                        "in": "header",
                        "name": "Authorization",
                        "description": "API Token / API Key for authentication",
                    }
                case AuthModel.BASIC:
                    security_schemas[security_schema_name] = {
                        "type": "http",
                    }
                case AuthModel.JWT:
                    security_schemas[security_schema_name] = {
                        "type": "apiKey",
                        "in": "header",
                        "name": "Authorization",
                        "description": "JWT Token for authentication",
                    }
                case AuthModel.SERVICE_ACCOUNT:
                    # TODO: this might not be 100% accurate
                    security_schemas[security_schema_name] = {
                        "type": "apiKey",
                        "in": "header",
                        "name": "Authorization",
                        "description": "Service Account Token for authentication",
                    }
                case _:
                    assert_never(auth_type)

        return authentication, authorization, security_schemas

    def get_oauth_scopes(self, credential_id: str, args: AppInfoRequest) -> dict[str, str]:
        oauth_scopes: dict[str, str] = {}

        for credential in self.integration.credentials:
            if credential.id == credential_id:
                if isinstance(credential.oauth_settings, OAuthConfig) and isinstance(
                    credential.oauth_settings, OAuthSettings
                ):
                    scopes = credential.oauth_settings.scopes

                    if scopes is None:
                        continue

                    if callable(scopes):
                        scopes = scopes(args)

                    oauth_scopes = {
                        scope: f"Required for {capability}"
                        for capability, scope in scopes.items()
                        if scope is not None
                    }

                break

        return oauth_scopes

    """
    Utilities
    """

    def _collect_refs(self, schema: dict[str, Any], refs: set[str]) -> None:
        if not isinstance(schema, dict):
            return

        if "$ref" in schema:
            ref_name = schema["$ref"].split("/")[-1]
            refs.add(ref_name)

        # Recursively process nested objects and arrays
        for value in schema.values():
            if isinstance(value, dict):
                self._collect_refs(value, refs)
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self._collect_refs(item, refs)

    def _create_base_spec(self) -> dict[str, Any]:
        """
        Creates the base specification for the app.
        """

        # App instance
        app = self.integration
        entitlement_types = app.entitlement_types
        resource_types = app.resource_types

        return {
            "openapi": "3.0.0",
            "info": {
                "x-app-id": app.app_id,
                "x-app-logo-url": app.description_data.logo_url,
                "x-app-vendor-domain": app.description_data.app_vendor_domain,
                "title": app.description_data.user_friendly_name,
                "description": app.description_data.description,
                "version": app.version,
                "x-entitlement-types": [
                    entitlement_type.model_dump() for entitlement_type in entitlement_types
                ],
                "x-resource-types": [
                    resource_type.model_dump() for resource_type in resource_types
                ],
                "x-categories": {
                    "type": "enum",
                    "enum": [category.value for category in app.description_data.categories],
                },
            },
            "servers": [],
            "paths": {},
            "components": {"securitySchemes": {}, "schemas": {}},
            "tags": [
                {
                    "name": "Read Capabilities",
                    "description": "These capabilities only perform read operations.",
                },
                {
                    "name": "Write Capabilities",
                    "description": "These capabilities perform write operations on top of potential read operations.",
                },
                {
                    "name": "Authorization",
                    "description": "OAuth 2.0 authorization capabilities.",
                },
                {
                    "name": "Specification",
                    "description": "Capabilities used for specifying the connector's capabilities.",
                },
            ],
        }

    def _replace_entitlement_types(self, spec: dict[str, Any]) -> None:
        """
        Replaces the entitlement types in the spec with the entitlement types from the integration.
        Depends on the x-entitlement-type extension to be set on the property level.
        """
        spec["properties"]["entitlement_type"] = {
            "type": "string",
            "title": spec["properties"]["entitlement_type"]["title"],
            "description": spec["properties"]["entitlement_type"]["description"],
            "enum": [
                entitlement_type.type_id for entitlement_type in self.integration.entitlement_types
            ],
        }

    def _replace_resource_types(self, spec: dict[str, Any]) -> None:
        """
        Replaces the resource types in the spec with the resource types from the integration.
        Depends on the x-resource-type extension to be set on the property level.
        """
        spec["properties"]["resource_type"] = {
            "type": "string",
            "title": spec["properties"]["resource_type"]["title"],
            "description": spec["properties"]["resource_type"]["description"],
            "enum": [resource_type.type_id for resource_type in self.integration.resource_types],
        }

    def _convert_null_type(self, schema: dict[str, Any]) -> dict[str, Any]:
        """Convert types with null type to nullable recursively."""
        if not isinstance(schema, dict):
            return schema

        # Handle single-value enums (do we even use those? what is TokenType for?)
        if "const" in schema:
            schema["enum"] = [schema.pop("const")]
            return schema

        # Handle anyOf definitions
        if "anyOf" in schema:
            types = [item for item in schema["anyOf"] if item.get("type") != "null"]
            has_null = len(types) != len(schema["anyOf"])

            if has_null and len(types) == 1:
                result = types[0]
                result["nullable"] = True
                return result

        # Recursively process all objects, sometimes this is more nested
        for key, value in schema.items():
            if isinstance(value, dict):
                schema[key] = self._convert_null_type(value)
            elif isinstance(value, list):
                schema[key] = [
                    self._convert_null_type(item) if isinstance(item, dict) else item
                    for item in value
                ]

        return schema

    def _dump_credentials(self) -> list[dict[str, Any]]:
        credentials_dump = []
        for cred in self.integration.credentials:
            cred_dict = cred.model_dump()
            if "input_model" in cred_dict:
                del cred_dict["input_model"]
            credentials_dump.append(cred_dict)
        return credentials_dump

    def _get_model_extended_json_schema(self, model: type[BaseModel]) -> dict[str, Any]:
        json_schema = model.model_json_schema(ref_template="#/components/schemas/{model}")
        field_order = list(model.model_fields.keys())
        json_schema["x-field_order"] = field_order
        return json_schema

    # Temp potentially
    def _serialize_for_json(self, obj: Any) -> Any:
        """Serialize Python objects for JSON output, handling special cases."""
        if obj is None:
            return None
        elif isinstance(obj, str | int | float | bool):
            return obj
        elif isinstance(obj, list | tuple | set):
            return [self._serialize_for_json(item) for item in obj]
        elif isinstance(obj, dict):
            return {
                key: self._serialize_for_json(value)
                for key, value in obj.items()
                if value is not None  # Skip None values
            }
        elif hasattr(obj, "model_dump"):  # Pydantic models
            return self._serialize_for_json(obj.model_dump())
        elif hasattr(obj, "__dict__"):  # Other objects with __dict__
            return self._serialize_for_json(obj.__dict__)
        else:
            return str(obj)  # Fallback to string representation

    def _process_definitions(self, obj: dict[str, Any] | list[Any], spec: dict[str, Any]) -> None:
        if isinstance(obj, dict):
            if "$defs" in obj:
                for def_name, def_schema in obj.pop("$defs").items():
                    if def_name not in spec["components"]["schemas"]:
                        spec["components"]["schemas"][def_name] = def_schema

                    # Perform some replace operations based on extensions
                    if any(
                        "x-entitlement-type" in property_schema
                        for property_schema in def_schema.get("properties", {}).values()
                    ):
                        self._replace_entitlement_types(def_schema)

                    if any(
                        "x-resource-type" in property_schema
                        for property_schema in def_schema.get("properties", {}).values()
                    ):
                        self._replace_resource_types(def_schema)

            if "$ref" in obj and obj["$ref"].startswith("#/$defs"):
                obj["$ref"] = "#/components/schemas/" + obj["$ref"].split("/")[-1]

        # Recursively process nested dictionaries and lists
        if isinstance(obj, dict):
            for value in obj.values():
                self._process_definitions(value, spec)

        elif isinstance(obj, list):
            for item in obj:
                if isinstance(item, dict):
                    self._process_definitions(item, spec)

    def _get_auth_type(self, auth_info: dict[str, Any]) -> str:
        credential_type = auth_info.get("x-credential-type", "unknown")

        if credential_type != "unknown":
            return AuthModel(credential_type).value

        return credential_type

    def _add_settings_to_schema(
        self, spec: dict[str, Any], model: type[BaseModel], used_refs: set[str]
    ):
        name = model.__name__
        if not hasattr(model, "model_json_schema"):
            raise ValueError(f"Model {name} does not have a model_json_schema method")

        schema = self._get_model_extended_json_schema(model)
        spec["components"]["schemas"]["Settings"] = schema
        used_refs.add("Settings")

        # Clean up
        if "$defs" in schema:
            del schema["$defs"]

        # Add enum schemas
        for field_name, field in model.model_fields.items():
            annotation = field.annotation
            if isinstance(annotation, type) and issubclass(annotation, Enum):
                enum_schema = {
                    "type": "string",
                    "enum": [e.value for e in annotation],
                    "title": annotation.__name__,
                    "description": annotation.__doc__,
                }
                spec["components"]["schemas"][annotation.__name__] = enum_schema
                # Update the reference in the model schema
                if "properties" in schema:
                    schema["properties"][field_name][
                        "$ref"
                    ] = f"#/components/schemas/{annotation.__name__}"
                    used_refs.add(annotation.__name__)

        # Add other referenced schemas
        for ref_name, ref_schema in schema.get("definitions", {}).items():
            if ref_name not in spec["components"]["schemas"]:
                spec["components"]["schemas"][ref_name] = ref_schema
                used_refs.add(ref_name)
        for field_name, field in schema.get("properties", {}).items():
            if isinstance(field, dict) and "$ref" in field:
                ref_key = field["$ref"].split("/")[-1]  # Extract schema name from $ref
                if ref_key not in spec["components"]["schemas"]:
                    annotation = model.model_fields[field_name].annotation
                    if annotation and hasattr(annotation, "model_json_schema"):
                        nested_schema = annotation.model_json_schema(
                            ref_template="#/components/schemas/{model}"
                        )
                        spec["components"]["schemas"][ref_key] = nested_schema
                        used_refs.add(ref_key)



import os
from enum import Enum
from pathlib import Path


class LogLevel(Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


LOG_LEVEL_ENV_VAR = "LUMOS_LOG_LEVEL"
LOG_DIRECTORY_ENV_VAR = "LUMOS_LOG_DIRECTORY"
LOG_TO_STDOUT_ENV_VAR = "LUMOS_LOG_TO_STDOUT"
ADDITIONAL_REDACTED_LOG_KEYS_ENV_VAR = "LUMOS_ADDITIONAL_REDACTED_LOG_KEYS"
LUMOS_ON_PREMISE_AGENT_SEND_CRITICAL_LOGS_TO_LUMOS_ENV_VAR = (
    "LUMOS_ON_PREMISE_AGENT_SEND_CRITICAL_LOGS_TO_LUMOS"
)
LUMOS_ON_PREMISE_AGENT_API_KEY_ENV_VAR = "LUMOS_ON_PREMISE_AGENT_API_KEY"
LUMOS_AGENT_IDENTIFIER_ENV_VAR = "LUMOS_AGENT_IDENTIFIER"
LUMOS_ON_PREMISE_AGENT_BASE_URL_ENV_VAR = "LUMOS_ON_PREMISE_AGENT_BASE_URL"


class Config:
    def __init__(self) -> None:
        self.log_level: LogLevel = LogLevel.ERROR
        self.log_directory: Path | None = None
        self.log_to_stdout: bool = False
        self.additional_redacted_log_keys: list[str] = []
        self.send_logs_to_on_prem_proxy: bool = False
        self.on_prem_agent_api_key: str | None = None
        self.agent_identifier: str | None = None
        self.base_url: str | None = None
        self.set_log_level()
        self.set_log_directory()
        self.set_log_to_stdout()
        self.set_send_logs_to_on_prem_proxy()
        self.set_additional_redacted_log_keys()
        self.set_on_prem_agent_api_key()
        self.set_agent_identifier()
        self.set_base_url()

    def set_log_level(self) -> None:
        log_level_string = os.environ.get(LOG_LEVEL_ENV_VAR, "ERROR")
        try:
            self.log_level = LogLevel(log_level_string)
        except ValueError as e:
            raise ValueError(
                f"Invalid log level set by environment variable {LOG_LEVEL_ENV_VAR}: {log_level_string}"
            ) from e

    def set_log_directory(self) -> None:
        log_directory_string = os.environ.get(LOG_DIRECTORY_ENV_VAR, None)
        if log_directory_string is not None:
            self.log_directory = Path(log_directory_string)

    def set_log_to_stdout(self) -> None:
        log_to_stdout_string = os.environ.get(LOG_TO_STDOUT_ENV_VAR, "False")
        self.log_to_stdout = log_to_stdout_string.lower() == "true"

    def set_additional_redacted_log_keys(self) -> None:
        additional_redacted_log_keys_string = os.environ.get(
            ADDITIONAL_REDACTED_LOG_KEYS_ENV_VAR, ""
        )
        self.additional_redacted_log_keys = [
            key.strip().lower() for key in additional_redacted_log_keys_string.split(",")
        ]

    def set_send_logs_to_on_prem_proxy(self) -> None:
        send_logs_to_on_prem_proxy_string = os.environ.get(
            LUMOS_ON_PREMISE_AGENT_SEND_CRITICAL_LOGS_TO_LUMOS_ENV_VAR, "False"
        )
        self.send_logs_to_on_prem_proxy = send_logs_to_on_prem_proxy_string.lower() == "true"

    def set_on_prem_agent_api_key(self) -> None:
        on_prem_agent_api_key_string = os.environ.get(LUMOS_ON_PREMISE_AGENT_API_KEY_ENV_VAR, None)
        if on_prem_agent_api_key_string is not None:
            self.on_prem_agent_api_key = on_prem_agent_api_key_string

    def set_agent_identifier(self) -> None:
        agent_identifier_string = os.environ.get(LUMOS_AGENT_IDENTIFIER_ENV_VAR, None)
        if agent_identifier_string is not None:
            self.agent_identifier = agent_identifier_string

    def set_base_url(self) -> None:
        base_url_string = os.environ.get(LUMOS_ON_PREMISE_AGENT_BASE_URL_ENV_VAR, None)
        if base_url_string is not None:
            self.base_url = base_url_string


config = Config()



from connector.client import get_page
from connector.generated import (
    FindEntitlementAssociationsRequest,
    FindEntitlementAssociationsResponse,
    FoundAccountData,
    GetLastActivityRequest,
    GetLastActivityResponse,
    ListAccountsRequest,
    ListAccountsResponse,
    ListCustomAttributesSchemaRequest,
    ListCustomAttributesSchemaResponse,
    ListEntitlementsRequest,
    ListEntitlementsResponse,
    ListResourcesRequest,
    ListResourcesResponse,
    Page,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)

from {name}.client import {pascal}Client
from {name}.pagination import DEFAULT_PAGE_SIZE, NextPageToken, Pagination, paginations_from_args

async def validate_credentials(
    args: ValidateCredentialsRequest,
) -> ValidateCredentialsResponse:
    async with {pascal}Client(args) as client:
        # _ = await client.get_users()

        return ValidateCredentialsResponse(
            response=ValidatedCredentials(
                unique_tenant_id="REPLACE_WITH_UNIQUE_TENANT_ID",
                valid=True,
            ),
        )


async def list_accounts(args: ListAccountsRequest) -> ListAccountsResponse:
    paginations, current_pagination, page_size = paginations_from_args(
        args, default_endpoints=["/users"]
    )

    async with {pascal}Client(args) as client:
        response = await client.get_users(
            limit=page_size,
            offset=current_pagination.offset,
        )
        accounts: list[FoundAccountData] = response.to_accounts()

        if True:
            paginations.append(
                Pagination(
                    endpoint=current_pagination.endpoint,
                    offset=current_pagination.offset + len(accounts),
                )
            )

    return ListAccountsResponse(
        response=accounts,
        page=NextPageToken.from_paginations(paginations).to_page(page_size)
        if paginations
        else None,
    )


async def list_resources(args: ListResourcesRequest) -> ListResourcesResponse:
    raise NotImplementedError


async def list_entitlements(
    args: ListEntitlementsRequest,
) -> ListEntitlementsResponse:
    raise NotImplementedError


async def find_entitlement_associations(
    args: FindEntitlementAssociationsRequest,
) -> FindEntitlementAssociationsResponse:
    raise NotImplementedError


async def get_last_activity(args: GetLastActivityRequest) -> GetLastActivityResponse:
    raise NotImplementedError


async def list_custom_attributes_schema(
    args: ListCustomAttributesSchemaRequest,
) -> ListCustomAttributesSchemaResponse:
    raise NotImplementedError



from datetime import datetime, timedelta, timezone

import jwt
from connector_sdk_types.generated import JWTCredential


def sign_jwt(credentials: JWTCredential, expiration_minutes: int = 20) -> str:
    # modify the claims to include the current `iat` and `exp` expiration time in UNIX time (seconds since the Unix epoch)
    now = datetime.now(timezone.utc)
    expiration_time = now + timedelta(minutes=expiration_minutes)
    credentials.claims.iat = int(now.timestamp())
    credentials.claims.exp = int(expiration_time.timestamp())
    token = jwt.encode(
        payload=credentials.claims.to_dict(),
        key=credentials.secret,
        headers=credentials.headers.to_dict(),
    )
    return token
